from typing import Callable, ClassVar, Literal

from orwynn.helpers.web import REQUEST_METHOD_BY_PROTOCOL, RequestMethod

from .endpoint.endpoint import Endpoint
from orwynn.base.controller.controller import Controller
from orwynn.utils.scheme import Scheme
from orwynn.http.errors import UnsupportedHttpMethodError
from .errors import \
    DefinedTwiceControllerMethodError
from orwynn.base.controller.errors import \
    MissingControllerClassAttributeError
from .endpoint.container import EndpointContainer
from orwynn.utils import validation


class HttpController(Controller):
    """Handles incoming requests and returns responses to the client.

    Calls service or different services under the hood to prepare a response.

    Controller is not a Provider, and it has lower priority than Service which
    means services can be easily injected into controller to be used.

    Controller should be assigned at according Module.controllers field to be
    initialized and registered.

    Class-Attributes:
        Route:
            A subroute of Module's route (where controller attached) which
            controller will answer to. This attribute is required to be
            defined in subclasses explicitly or an error will be raised.
            It is allowed to be "/" to handle Module's root route requests.
        Endpoints:
            List of endpoints enabled in this controller.
        Version (optional):
            An API version the controller supports. Defaults to the latest one.
    """
    Route: ClassVar[str | None] = None
    Endpoints: ClassVar[list[Endpoint] | None] = None
    Version: ClassVar[int | set[int] | Literal["*"] | None] = None

    def __init__(self) -> None:
        super().__init__()

        self._methods: list[RequestMethod] = []

        if self.Endpoints is None:
            raise MissingControllerClassAttributeError(
                "you should set class attribute Endpoints for"
                f" controller {self.__class__}"
            )
        else:
            validation.validate_each(
                self.Endpoints,
                Endpoint,
                expected_sequence_type=list,
                should_check_if_empty=True
            )
            collected_str_methods: list[str] = []
            for endpoint in self.Endpoints:
                str_method = endpoint.method.lower()

                http_methods: list[RequestMethod] = [
                    method.value
                    for method in REQUEST_METHOD_BY_PROTOCOL[Scheme.HTTP]
                ]
                if str_method not in http_methods:
                    raise UnsupportedHttpMethodError(
                        f"method {str_method} is not supported"
                    )
                if str_method in collected_str_methods:
                    raise DefinedTwiceControllerMethodError(
                        f"method {str_method} is defined twice for"
                        f" controller {self.__class__}"
                    )

                collected_str_methods.append(str_method)
                http_method: RequestMethod = RequestMethod(str_method)

                if (
                    http_method
                    not in REQUEST_METHOD_BY_PROTOCOL[Scheme.HTTP]
                ):
                    raise ValueError(f"cannot accept method {http_method}")

                self._methods.append(http_method)

                EndpointContainer.ie().add(
                    self.get_fn_by_http_method(http_method),
                    endpoint
                )

    @property
    def route(self) -> str:
        return self._route

    @property
    def methods(self) -> list[RequestMethod]:
        return self._methods

    def get_fn_by_http_method(self, method: RequestMethod) -> Callable:
        fn: Callable

        match method:
            case RequestMethod.GET:
                fn = self.get
            case RequestMethod.POST:
                fn = self.post
            case RequestMethod.PUT:
                fn = self.put
            case RequestMethod.DELETE:
                fn = self.delete
            case RequestMethod.PATCH:
                fn = self.patch
            case RequestMethod.OPTIONS:
                fn = self.options
            case _:
                raise

        return fn

    def get(self, *args, **kwargs) -> dict:
        raise NotImplementedError(
            "the method GET is not implemented for controller"
            f" {self.__class__}"
        )

    def post(self, *args, **kwargs) -> dict:
        raise NotImplementedError(
            "the method POST is not implemented for controller"
            f" {self.__class__}"
        )

    def put(self, *args, **kwargs) -> dict:
        raise NotImplementedError(
            "the method PUT is not implemented for controller"
            f" {self.__class__}"
        )

    def delete(self, *args, **kwargs) -> dict:
        raise NotImplementedError(
            "the method DELETE is not implemented for controller"
            f" {self.__class__}"
        )

    def patch(self, *args, **kwargs) -> dict:
        raise NotImplementedError(
            "the method PATCH is not implemented for controller"
            f" {self.__class__}"
        )

    def options(self, *args, **kwargs) -> dict:
        raise NotImplementedError(
            "the method OPTIONS is not implemented for controller"
            f" {self.__class__}"
        )



import functools
from typing import Callable

from orwynn.base.error.errors import MalfunctionError
from orwynn.utils import url
from orwynn.websocket.websocket import Websocket

from .handlers import DispatchWebsocketHandler, WebsocketHandler
from .nextcall import NextCallHandler


class WebsocketStack:
    """
    Manages an order of calls on websocket's request way through middleware and
    controllers.

    Call stack is built from handler functions and their routes to respond to.
    These functions called one by one for each request according to the order
    they were inserted.

    These functions may be either WebsocketMiddleware.dispatch methods or
    certain WebsocketController methods.

    Attributes:
        handler_register:
            Registration entity for websocket handlers.
    """
    def __init__(
        self,
        handler_register: Callable
    ) -> None:
        self.__handler_register: Callable = handler_register
        self.__handlers_by_route: dict[str, list[WebsocketHandler]] = {}
        self.__is_registered: bool = False

    def add_call(
        self,
        handler: WebsocketHandler
    ) -> None:
        """
        Adds a call to the stack.

        Note that handler's route can be wildcard "*", i.e. this handler will
        process all routes, and it should be Middleware-based (have a dispatch
        function). The wildcarded Middleware always called before other route
        handlers.

        Also there is an error raised, if a dispatch handler is placed
        after non-dispatch handler for any route provided.

        Args:
            handler:
                Webscoket handler object to add.
        """
        if self.__is_registered:
            raise ValueError(
                "the stack has been already registered"
            )

        if handler.route not in self.__handlers_by_route:
            self.__handlers_by_route[handler.route] = []
        # Here we should be sure that the list for the route is not only
        # initialized, but also is populated with at least one value
        elif len(self.__handlers_by_route[handler.route]) < 1:
            raise MalfunctionError(
                f"handler list for route {handler.route} is initialized"
                " but no values are added, but should be at least one"
            )
        elif (
            isinstance(handler, DispatchWebsocketHandler)
            # Check only last handler (i.e. previously added) in the list
            and not isinstance(
                self.__handlers_by_route[handler.route][-1],
                DispatchWebsocketHandler
            )
        ):
            raise ValueError(
                f"cannot insert dispatch handler {handler} for route"
                f" {handler.route}, since previously non-dispatch handler"
                f" {self.__handlers_by_route[handler.route][-1]}"
                " has been already inserted"
            )

        self.__handlers_by_route[handler.route].append(handler)

    def register_all(
        self
    ) -> None:
        """
        Registers entrypoints to the websocket handlers.

        After calling of this method no more adds are allowed.
        """
        # Check wildcard route consists of only dispatch handlers
        if not all([
            isinstance(handler, DispatchWebsocketHandler)
            for handler in self.__handlers_by_route["*"]
        ]):
            raise ValueError(
                "wildcard handlers have non-dispatch based handler"
            )

        # For every route register the entry Connection middleware.
        for route, handlers in self.__handlers_by_route.items():
            # Skip wildcard routes since they cannot contain final generic
            # handler
            if route == "*":
                continue

            # Non-wildcard route handlers shouldn't contain dispatch handlers
            if not all([
                not isinstance(handler, DispatchWebsocketHandler)
                for handler in handlers
            ]):
                raise ValueError(
                    f"non-wildcard route {route} handlers shouldn't"
                    " contain dispatch handlers"
                )

            # List of final handlers to be registered should consist of
            # wildcard handlers first, and only then of other generic ones.
            final_handlers: list[WebsocketHandler] = \
                self.__handlers_by_route["*"] + handlers
            self.__handler_register(route)(functools.partial(
                self.__handle_entry,
                _fw_handlers=final_handlers,
                _fw_original_route=route
            ))
        self.__is_registered = True

    async def __handle_entry(
        self,
        websocket: Websocket,
        *,
        _fw_handlers: list[WebsocketHandler],
        _fw_original_route: str
    ) -> None:
        # Execute all handlers consequently, but delegate such execution inside
        # the registered middleware
        first_handler: WebsocketHandler = _fw_handlers[0]

        # Collect request kwargs from url
        url_vars: url.UrlVars = url.get_vars(
            websocket.url,
            abstract_route=_fw_original_route
        )

        # For middleware pass special NextCallHandler with data of all handler
        # functions ahead, except first one which is executed first.
        if isinstance(first_handler, DispatchWebsocketHandler):
            await first_handler.fn(
                websocket,
                NextCallHandler(_fw_handlers[1:], url_vars=url_vars)
            )
        # For all other generic cases, we might just call with only websocket
        # passed. But this situation rarely possible since we have added
        # BuiltinMiddleware (at least one), so here we will raise an error.
        else:
            raise MalfunctionError(
                "expected middleware as the first handler (due to builtin"
                " middlewares adding by default), but got handler"
                f" {first_handler} instead"
            )



import pytest

from orwynn.base.module.module import Module
from orwynn.boot.boot import Boot
from orwynn.http import Endpoint, HttpController
from orwynn.utils import validation

from . import ApiVersion
from .errors import UnsupportedVersionError

# NOTE: By default there is no global route for backwards compatiblity.

@pytest.mark.asyncio
async def test_versioned_global_route():
    class C(HttpController):
        Route = "/message"
        Endpoints = [Endpoint(method="get")]

        def get(self) -> dict:
            return {"message": "hello"}

    boot: Boot = await Boot.create(
        root_module=Module("/user", Controllers=[C]),
        global_http_route="/donuts/v{version}",
    )

    boot.app.client.get_jsonify(
        "/donuts/v1/user/message",
        200,
        is_global_route_used=False
    )


@pytest.mark.asyncio
async def test_controller_version():
    """
    HttpController can define older version of API than available.
    """
    class C1(HttpController):
        Route = "/message"
        Version = 1
        Endpoints = [Endpoint(method="get")]

        def get(self) -> dict:
            return {"message": "hello v1"}

    class C2(HttpController):
        # Here we don't need to define a Version, since the v2 should be
        # latest.
        Route = "/message"
        Endpoints = [Endpoint(method="get")]

        def get(self) -> dict:
            return {"message": "hello v2"}

    boot: Boot = await Boot.create(
        root_module=Module("/user", Controllers=[C1, C2]),
        global_http_route="/api/v{version}",
        api_version=ApiVersion(
            supported={1, 2}
        )
    )

    data: dict

    data = boot.app.client.get_jsonify(
        "/user/message", 200, api_version=1
    )
    assert data["message"] == "hello v1"

    data = boot.app.client.get_jsonify(
        "/user/message", 200, api_version=2
    )
    assert data["message"] == "hello v2"

    data = boot.app.client.get_jsonify(
        "/api/v3/user/message", 404, is_global_route_used=False
    )


@pytest.mark.asyncio
async def test_controller_all_versions():
    class C1(HttpController):
        Route = "/message"
        Version = "*"
        Endpoints = [Endpoint(method="get")]

        def get(self) -> dict:
            return {"message": "hello"}

    boot: Boot = await Boot.create(
        root_module=Module("/user", Controllers=[C1]),
        global_http_route="/api/v{version}",
        api_version=ApiVersion(
            supported={1, 2, 3}
        )
    )

    data: dict

    data = boot.app.client.get_jsonify("/user/message", 200, api_version=1)
    assert data["message"] == "hello"

    data = boot.app.client.get_jsonify("/user/message", 200, api_version=2)
    assert data["message"] == "hello"

    data = boot.app.client.get_jsonify("/user/message", 200, api_version=3)
    assert data["message"] == "hello"


@pytest.mark.asyncio
async def test_controller_several_versions():
    class C1(HttpController):
        Route = "/message"
        Version = {2, 3}
        Endpoints = [Endpoint(method="get")]

        def get(self) -> dict:
            return {"message": "hello"}

    boot: Boot = await Boot.create(
        root_module=Module("/user", Controllers=[C1]),
        global_http_route="/api/v{version}",
        api_version=ApiVersion(
            supported={1, 2, 3}
        )
    )

    data: dict

    data = boot.app.client.get_jsonify("/user/message", 404, api_version=1)

    data = boot.app.client.get_jsonify("/user/message", 200, api_version=2)
    assert data["message"] == "hello"

    data = boot.app.client.get_jsonify("/user/message", 200, api_version=3)
    assert data["message"] == "hello"


@pytest.mark.asyncio
async def test_controller_unsupported_version():
    class C1(HttpController):
        Route = "/message"
        Version = 3
        Endpoints = [Endpoint(method="get")]

        def get(self) -> dict:
            return {"message": "hello"}

    await validation.expect_async(
        Boot.create(
            root_module=Module("/user", Controllers=[C1]),
            global_http_route="/api/v{version}",
            api_version=ApiVersion(
                supported={1, 2}
            )
        ),
        UnsupportedVersionError
    )


@pytest.mark.asyncio
async def test_controller_unsupported_version_of_many():
    class C1(HttpController):
        Route = "/message"
        # Some are supported, some are not
        Version = {2, 3}
        Endpoints = [Endpoint(method="get")]

        def get(self) -> dict:
            return {"message": "hello"}

    await validation.expect_async(
        Boot.create(
            root_module=Module("/user", Controllers=[C1]),
            global_http_route="/api/v{version}",
            api_version=ApiVersion(
                supported={1, 2}
            )
        ),
        UnsupportedVersionError
    )



from enum import Enum


class Indicator(Enum):
    """List of indicators applicable to Schema's data representation.

    Values:
        TYPE:
            Indicates that field should contain any type representation of an
            object.
        VALUE:
            Indicates that field should contain a value of an object.
    """
    TYPE = "type"
    VALUE = "value"



from orwynn.base.module import Module
from orwynn.mongo.search import DocumentSearch, MongoStateFlagSearch
from orwynn.mongo.services import MongoStateFlagService
from orwynn.mongo.stateflag import MongoStateFlag
from orwynn.mongo.utils import MongoUtils

from .config import MongoConfig
from .document import Document
from .mongo import Mongo

__all__ = [
    "Document",
    "MongoConfig",
    "Mongo",
    "MongoUtils",
    "DocumentSearch",
    "MongoStateFlag",
    "MongoStateFlagSearch",
    "MongoStateFlagService"
]

module = Module(
    Providers=[Mongo, MongoConfig, MongoStateFlagService],
    exports=[Mongo, MongoConfig, MongoStateFlagService]
)



from orwynn.context import ContextStorage
from orwynn.context.errors import RequestIdAlreadySavedError
from orwynn.utils import validation
from orwynn.utils.rnd import makeid


class HttpRequestContextId:
    def __init__(self) -> None:
        self.__storage: ContextStorage = ContextStorage.ie()

    def get(
        self
    ) -> str:
        return validation.apply(self.__storage.get("request_id"), str)

    def save(self) -> str:
        """Generates an id for the request and saves it into the context.

        Returns:
            Request id generated.

        Raises:
            RequestIdAlreadySavedError:
                If the request id has been set previously.
        """
        request_id: str = makeid()
        try:
            self.__storage.get("request_id")
        except KeyError:
            self.__storage.save("request_id", request_id)
            return request_id
        else:
            raise RequestIdAlreadySavedError()



import pytest

from orwynn.utils.klass import ClassUtils


def test_bind_first_arg():
    class A:
        def hello(self, x: str, y: int) -> str:
            return x * y

    a = A()
    assert ClassUtils.bind_first_arg(a)(A.hello)("wow", 4) == "wowwowwowwow"


@pytest.mark.asyncio
async def test_bind_first_arg_async():
    class A:
        async def hello(self, x: str, y: int) -> str:
            return x * y

    a = A()
    assert await ClassUtils.bind_first_arg_async(a)(A.hello)(
        "wow", 4
    ) == "wowwowwowwow"



import pytest
import pytest_asyncio

from orwynn import mongo
from orwynn.base.model.model import Model
from orwynn.base.module.module import Module
from orwynn.boot.boot import Boot
from orwynn.http import Endpoint, EndpointResponse, HttpController
from orwynn.router.errors import (
    UnmatchedEndpointEntityError,
)
from orwynn.utils import validation


@pytest_asyncio.fixture
async def run_endpoint():
    class Item(Model):
        name: str
        price: int

    class Response400(Model):
        message: str
        power: float

    class C1(HttpController):
        Route = "/"
        Endpoints = [
            Endpoint(
                method="get",
                default_status_code=201,
                summary="Best summary",
                tags=["best-item", "buy-now"],
                is_deprecated=True,
                responses=[
                    EndpointResponse(
                        status_code=201,
                        Entity=Item,
                        description="Best response description"
                    ),
                    EndpointResponse(
                        status_code=400,
                        Entity=Response400
                    )
                ]
            )
        ]

        def get(self) -> Item:
            return Item(name="hello", price=1)

    await Boot.create(
        Module(
            route="/",
            Controllers=[C1],
            imports=[mongo.module]
        ),
        apprc={
            "test": {
                "Mongo": {
                    "url": "mongodb://localhost:9006",
                    "database_name": "orwynn_test"
                }
            }
        }
    )


@pytest.mark.asyncio
async def test_not_matched_spec_to_return_type():
    class WrongItem(Model):
        name: str
        doubt: int

    class Item(Model):
        name: str
        price: float

    class C1(HttpController):
        Route = "/"
        Endpoints = [
            Endpoint(
                method="get",
                responses=[EndpointResponse(status_code=200, Entity=WrongItem)]
            )
        ]

        def get(self) -> Item:
            return Item(name="hello", price=1.2)

    await validation.expect_async(
        Boot.create(
            Module(
                route="/",
                Controllers=[C1]
            )
        ),
        UnmatchedEndpointEntityError
    )



import typing
from typing import Callable

from starlette.types import Receive, Scope, Send

from orwynn.app.config import AppConfig
from orwynn.app.core import CoreApp
from orwynn.app.types import CoreCors
from orwynn.base.service.framework import FrameworkService
from orwynn.helpers.web import RequestMethod
from orwynn.testing import Client, EmbeddedTestClient


class App(FrameworkService):
    def __init__(self, config: AppConfig) -> None:
        self._core_app: CoreApp = CoreApp(
            docs_url=config.docs_route,
            redoc_url=config.redoc_route
        )
        self._wrapped_core_app: CoreApp
        if config.cors and config.cors.is_enabled:
            # cors should wrap an application to add required headers to error
            # responses too
            self._wrapped_core_app = typing.cast(
                CoreApp,
                CoreCors(
                    app=self._core_app,
                    allow_origins=config.cors.allow_origins or (),
                    allow_methods=config.cors.allow_methods or ("GET",),
                    allow_headers=config.cors.allow_headers or (),
                    allow_credentials=config.cors.allow_credentials,
                    allow_origin_regex=config.cors.allow_origin_regex,
                    expose_headers=config.cors.expose_headers or (),
                    max_age=config.cors.max_age
                )
            )
        else:
            self._wrapped_core_app = self._core_app

        self._fw_register_middleware = self._core_app.add_middleware
        self._fw_register_exception_handler_fn = \
            self._core_app.add_exception_handler
        self._fw_websocket_handler = self._core_app.websocket

        self.HTTP_METHODS_TO_REGISTERING_FUNCTIONS: \
            dict[RequestMethod, Callable] = {
                RequestMethod.GET: self._core_app.get,
                RequestMethod.POST: self._core_app.post,
                RequestMethod.PUT: self._core_app.put,
                RequestMethod.DELETE: self._core_app.delete,
                RequestMethod.PATCH: self._core_app.patch,
                RequestMethod.OPTIONS: self._core_app.options
            }

        self._client: Client = Client(
            EmbeddedTestClient(self._wrapped_core_app)
        )

    async def __call__(
        self, scope: Scope, receive: Receive, send: Send
    ) -> None:
        await self._wrapped_core_app(scope, receive, send)

    @property
    def core_app(self) -> CoreApp:
        return self._core_app

    @property
    def client(self) -> Client:
        return self._client



from sqlalchemy.orm import Mapped, mapped_column

from orwynn.sql.table import Table


class StateFlag(Table):
    """
    Holds some flag signifies state of database.
    """
    name: Mapped[str] = mapped_column(unique=True)
    value: Mapped[bool]



from orwynn.mongo.document.testing import SimpleDocument


def test_create(mongo_boot):
    created = SimpleDocument(name="pizza", price=1.2).create()

    assert isinstance(created.getid(), str)
    assert created.name == "pizza"
    assert created.price == 1.2



"""Common types.
"""
from typing import Any, Callable, Coroutine, TypeVar

T = TypeVar("T")

TClass = TypeVar("TClass", bound=type)
TDecoratedCallable = TypeVar("TDecoratedCallable", bound=Callable[..., Any])

Timestamp = float
Delta = float

CashOperator = str
"""
Special string literal starts with dollar sign `$` which holds a special
meaning to the acceptor logic.
"""

AnyCoro = Coroutine[Any, Any, Any]



from typing import Any

from fastapi.exceptions import (
    RequestValidationError as FastAPIRequestValidationError,
)

RequestValidationException = FastAPIRequestValidationError


class ExpectationError(Exception):
    pass


class NothingToValidateError(Exception):
    """Typically raised if an empty structure is passed to validation function.
    """


class ReValidationError(Exception):
    def __init__(
        self,
        message: str = "",
        failed_obj: Any | None = None,
        pattern: str | None = None,
        *args
    ) -> None:
        if not message and failed_obj is not None and pattern is not None:
            message = \
                f"{repr(failed_obj)} should implement pattern {pattern}"

        super().__init__(message)


class UnknownValidatorError(Exception):
    pass


class ValidationError(Exception):
    def __init__(
        self,
        message: str = "",
        failed_obj: Any | None = None,
        expected_type: type | list[type] | None = None,
        *args
    ) -> None:
        if (
            not message
            and failed_obj is not None
            # No need to add "expected_type is not None" here since we
            # sometimes compare objects to None
        ):
            if isinstance(expected_type, type):
                message = \
                    f"{repr(failed_obj)} should have type:" \
                    f" {expected_type.__name__}"
            elif type(expected_type) is list:
                message = \
                    f"{repr(failed_obj)}" \
                    " should be any type of" \
                    + f" list: {[type_.__name__ for type_ in expected_type]}"
            else:
                raise TypeError("Unrecognized type of `expected_type`")

        super().__init__(message)



from typing import Any, Generic, TypeVar

from orwynn.utils.condition._mark import ComparisonMark
from orwynn.utils.condition.errors import UnsupportedComparisonError
from antievil import UnsupportedError, WrongGenericTypeError

T = TypeVar("T")


class ComparisonCondition(Generic[T]):
    """
    Contains condition mark and a value to compare to target.

    Attributes:
        mark:
            Compare mark relative to target. For example if the mark is set to
            MoreEqual, it means that the target should be more or equal than
            the value given to this condition, i.e.
            target >= ComparisonCondition.value.
        value:
            Value to compare to. Should be Comparable and contain all methods
            of comparison to satisfy ComparisonMark fields (aka python dunder
            comparators like `__le__`).
    """

    # TODO(ryzhovalex): ensure that operated value and target type
    #   can handle the chosen mark's comparison (i.e. has an according dunder
    #   comparator method)
    # 0

    def __init__(self, mark: ComparisonMark, value: T) -> None:
        self._mark: ComparisonMark = mark
        self._value: T = value

    def __str__(self) -> str:
        return f"\"{self.mark.value}\" {self.value}"

    @property
    def mark(self) -> ComparisonMark:
        return self._mark

    @property
    def value(self) -> T:
        return self._value

    def compare(self, target: T) -> bool:
        """
        Compares given target to the condition's value using condition's
        defined compare mark.

        Args:
            target:
                Target to compare.

        Returns:
            Comparison result as a boolean.

        Raises:
            UnsupportedComparisonError:
                Type of target and value does not support chosen comparison.
            WrongGenericTypeError:
                Target's type is different from condition's value type.
            UnsupportedError:
                Unrecognized compare mark defined by condition.
        """
        self._check_target_type(target)

        try:
            return self._compare(target)
        except TypeError as error:
            raise UnsupportedComparisonError(
                Type=type(target),
                compare_mark=self._mark,
            ) from error

    def _compare(
        self,
        target: T,
    ) -> bool:
        match self._mark:
            case ComparisonMark.Equal:
                return target == self._value
            case ComparisonMark.NotEqual:
                return target != self._value
            # ignore linter comparison errors since we handle it at
            # runtime
            case ComparisonMark.More:
                return target > self._value  # type: ignore
            case ComparisonMark.Less:
                return target < self._value  # type: ignore
            case ComparisonMark.MoreEqual:
                return target >= self._value  # type: ignore
            case ComparisonMark.LessEqual:
                return target <= self._value  # type: ignore
            case _:
                raise UnsupportedError(
                    title="compare mark",
                    value=self._mark,
                )

    def _check_target_type(self, target: T) -> None:
        if type(target) is not type(self._value):
            raise WrongGenericTypeError(
                GenericClass=type(self),
                ExpectedType=type(self._value),
                ReceivedType=type(target),
            )



from typing import Awaitable, Callable

from orwynn.websocket.websocket import Websocket

from .genericfn import GenericWebsocketFn

DispatchWebsocketFn = Callable[
    [Websocket, GenericWebsocketFn], Awaitable[None]
]



from orwynn.base.model.model import Model

from .dispatchfn import DispatchWebsocketFn
from .genericfn import GenericWebsocketFn


class WebsocketHandler(Model):
    fn: GenericWebsocketFn
    route: str


class DispatchWebsocketHandler(WebsocketHandler):
    fn: DispatchWebsocketFn
    # Middleware is responding to all routes, but inside own dispatch function
    # decides whether to handle it or not.
    route: str = "*"



import re
from itertools import zip_longest

from orwynn.utils.url.url import Url
from orwynn.utils.url.utils import match_routes
from orwynn.utils.url.vars import UrlVars


def get_vars(
    url: Url,
    *,
    abstract_route: str
) -> UrlVars:
    """
    Inspects the given url to find all path and query vars.

    Args:
        url:
            Url to inspect.
        abstract_route:
            Route with formatting blocks for route vars included.

    Returns:
        Dict formed of path and query vars.
    """
    final: UrlVars = UrlVars(
        url=url,
        abstract_route=abstract_route,
        path_vars={},
        query_vars={}
    )
    _set_path_vars(final)
    if url.query != "":
        _set_query_vars(final)
    return final


def _set_path_vars(
    url_vars: UrlVars
) -> None:
    path_var_names: list[str] = re.findall(
        r"\{\w+\}",
        url_vars.abstract_route
    )

    # Remove brackets from each element in acquired path variable names
    path_var_names = [
        name.replace("{", "").replace("}", "") for name in path_var_names
    ]

    path_match = match_routes(
        abstract_route=url_vars.abstract_route,
        real_route=url_vars.url.path
    )

    if path_match:
        for var_name, var_value in zip_longest(
            path_var_names, path_match.groups()
        ):
            if var_name is None or var_value is None:
                raise ValueError(
                    "unmatched amount of variable names and it's values in"
                    " given abstract and real routes"
                )

            url_vars.path_vars[var_name] = var_value


def _set_query_vars(
    url_vars: UrlVars
) -> None:
    query: str = url_vars.url.query
    query_elements: list[str] = query.split("&")
    for el in query_elements:
        if el == "":
            continue
        name, value = el.split("=")
        url_vars.query_vars[name] = value



import pytest

from orwynn.app import AppMode
from orwynn.apprc.apprc import AppRc
from orwynn.base.config import Config
from orwynn.base.model.model import Model
from orwynn.base.module.module import Module
from orwynn.boot import Boot
from orwynn.di.di import Di
from orwynn.utils import validation
from orwynn.utils.mp import find as mp_find


class Menu(Model):
    __root__: dict[str, float]


class BurgerShotConfig(Config):
    location: str
    employees: int
    menu: Menu


@pytest.fixture
def raw_apprc() -> AppRc:
    return {
        "prod": {
            "BurgerShot": {
                "location": "Vinewood",
                "employees": 15,
                "menu": {
                    "standard_burger": 1.25,
                    "cola": 1.5
                }
            }
        },
        "dev": {
            "BurgerShot": {
                "employees": 25,
                "menu": {
                    "cola": 1.8,
                    "pizza": 4.1,
                    "fried_chicken": 3.5
                }
            }
        },
        "test": {
            "BurgerShot": {
                "location": "Jefferson",
                "menu": {
                    "standard_burger": 1.1,
                    "fried_chicken": 5.0
                }
            }
        }
    }


@pytest.mark.asyncio
async def test_prod(raw_apprc: AppRc):
    await Boot.create(
        root_module=Module("/", Providers=[BurgerShotConfig]),
        apprc=raw_apprc,
        mode=AppMode.PROD
    )

    assert validation.apply(
        Di.ie().find("BurgerShotConfig"),
        BurgerShotConfig
    ).dict() == mp_find("prod.BurgerShot", raw_apprc)


@pytest.mark.asyncio
async def test_dev(raw_apprc: AppRc):
    await Boot.create(
        root_module=Module("/", Providers=[BurgerShotConfig]),
        apprc=raw_apprc,
        mode=AppMode.DEV
    )

    assert validation.apply(
        Di.ie().find("BurgerShotConfig"),
        BurgerShotConfig
    ).dict() == {
        "location": "Vinewood",
        "employees": 25,
        "menu": {
            "standard_burger": 1.25,
            "cola": 1.8,
            "pizza": 4.1,
            "fried_chicken": 3.5
        }
    }


@pytest.mark.asyncio
async def test_test(raw_apprc: AppRc):
    await Boot.create(
        root_module=Module("/", Providers=[BurgerShotConfig]),
        apprc=raw_apprc,
        mode=AppMode.TEST
    )

    assert validation.apply(
        Di.ie().find("BurgerShotConfig"),
        BurgerShotConfig
    ).dict() == {
        "location": "Jefferson",
        "employees": 25,
        "menu": {
            "standard_burger": 1.1,
            "cola": 1.8,
            "pizza": 4.1,
            "fried_chicken": 5.0
        }
    }



import re
from types import NoneType
from typing import ClassVar, Literal

from orwynn.base.controller.errors import (
    MissingControllerClassAttributeError,
)
from orwynn.utils import validation
from orwynn.utils.url import match_routes


class Controller:
    """
    Entrypoint to some operational service.
    """
    Route: ClassVar[str | None] = None
    Version: ClassVar[int | set[int] | Literal["*"] | None] = None

    def __init__(self) -> None:
        # Actual route for the controller assigned at boottime
        self._final_routes: set[str] = set()

        if self.Route is None:
            raise MissingControllerClassAttributeError(
                "you should set class attribute Route for"
                f" controller {self.__class__}"
            )
        else:
            validation.validate(self.Route, str)
            validation.validate_route(self.Route)
            self._route: str = self.Route

        validation.validate(self.Version, [int, set, str, NoneType])
        if isinstance(self.Version, str) and self.Version != "*":
            raise TypeError(
                f"unrecognized Version {self.Version}"
            )

    @property
    def final_routes(self) -> set[str]:
        if len(self._final_routes) == 0:
            raise ValueError(
                f"final routes is not set for controller {self}"
            )
        return self._final_routes.copy()

    @property
    def route(self) -> str:
        return self._route

    def is_matching_route(self, real_route: str) -> bool:
        """
        Checks if given real route is matching any of controller's final
        abstract routes.

        Args:
            real_route:
                Real route with all placement blocks filled.

        Returns:
            Boolean flag.
        """
        for abstract_route in self.final_routes:
            match_result: re.Match | None = match_routes(
                abstract_route=abstract_route,
                real_route=real_route
            )

            if match_result is not None:
                return True

        return False

    def _fw_update_final_routes(self, value: set[str]) -> None:
        self._final_routes.update(value)



from typing import Any

from antievil import NotFoundError

from orwynn.base import Service
from orwynn.mongo.search import MongoStateFlagSearch
from orwynn.mongo.stateflag import MongoStateFlag
from orwynn.mongo.utils import MongoUtils
from orwynn.utils.func import FuncSpec


class MongoStateFlagService(Service):
    def get(
        self,
        search: MongoStateFlagSearch
    ) -> list[MongoStateFlag]:
        query: dict[str, Any] = {}

        if search.ids:
            query["id"] = {
                "$in": search.ids
            }
        if search.keys:
            query["key"] = {
                "$in": search.keys
            }
        if search.values:
            query["value"] = {
                "$in": search.values
            }

        return MongoUtils.process_query(
            query,
            search,
            MongoStateFlag
        )

    def get_first_or_set_default(
        self,
        key: str,
        default_value: bool
    ) -> MongoStateFlag:
        """
        Returns first flag found for given search or a new flag initialized
        with default value.
        """
        try:
            return self.get(MongoStateFlagSearch(
                keys=[key]
            ))[0]
        except NotFoundError:
            return self.set(
                key,
                default_value
            )

    def set(
        self,
        key: str,
        value: bool
    ) -> MongoStateFlag:
        """
        Sets new value for a key.

        If the key does not exist, create a new state flag with given value.
        """
        flag: MongoStateFlag

        try:
            flag = self.get(MongoStateFlagSearch(
                keys=[key]
            ))[0]
        except NotFoundError:
            flag = MongoStateFlag(key=key, value=value).create()
        else:
            flag = flag.update(set={"value": value})

        return flag

    def decide(
        self,
        *,
        key: str,
        on_true: FuncSpec | None = None,
        on_false: FuncSpec | None = None,
        finally_set_to: bool,
        default_flag_on_not_found: bool
    ) -> Any:
        """
        Takes an action based on flag retrieved value.

        Args:
            key:
                Key of the flag to search for.
            finally_set_to:
                To which value the flag should be set after the operation is
                done.
            default_flag_on_not_found:
                Which value is to set for the unexistent by key flag.
            on_true(optional):
                Function to be called if the flag is True. Nothing is called
                by default.
            on_false(optional):
                Function to be called if the flag is False. Nothing is called
                by default.

        Returns:
            Chosen function output. None if no function is used.
        """
        result: Any = None
        flag: MongoStateFlag = self.get_first_or_set_default(
            key, default_flag_on_not_found
        )

        if flag.value is True and on_true is not None:
            result = on_true.call()
        if flag.value is False and on_false is not None:
            result = on_false.call()

        flag.update(
            set={
                "value": finally_set_to
            }
        )

        return result



import pytest

from orwynn.base.errorhandler.errorhandler import ErrorHandler
from orwynn.base.module.module import Module
from orwynn.boot.boot import Boot
from orwynn.proxy.boot import BootProxy
from orwynn.utils import validation
from orwynn.utils.scheme import Scheme
from orwynn.websocket import Websocket, WebsocketController


class SomeWebsocketError(Exception):
    pass


class Wc(WebsocketController):
    Route = "/"

    async def main(self, ws: Websocket) -> None:
        raise SomeWebsocketError("hello")


class Eh(ErrorHandler):
    E = SomeWebsocketError
    PROTOCOL = Scheme.WEBSOCKET

    async def handle(
        self, request: Websocket, error: SomeWebsocketError
    ) -> None:
        data: dict = BootProxy.ie().api_indication.digest(error)
        data["value"]["message"] = "handled"
        await request.send_json(data)


@pytest.mark.asyncio
async def test_default():
    boot: Boot = await Boot.create(
        Module("/", Controllers=[Wc])
    )

    with boot.app.client.websocket("/") as ws:
        data: dict = ws.receive_json()

        err = validation.apply(
            BootProxy.ie().api_indication.recover(
                SomeWebsocketError,
                data
            ),
            SomeWebsocketError
        )
        assert err.args[0] == "hello"


@pytest.mark.asyncio
async def test_one_handler():
    boot: Boot = await Boot.create(
        Module("/", Controllers=[Wc]),
        ErrorHandlers={Eh}
    )

    with boot.app.client.websocket("/") as ws:
        data: dict = ws.receive_json()

        err = validation.apply(
            BootProxy.ie().api_indication.recover(
                SomeWebsocketError,
                data
            ),
            SomeWebsocketError
        )
        assert err.args[0] == "handled"



from orwynn.mongo.document.testing import NestedDocument, SimpleDocument


def test_main(document_1: SimpleDocument):
    assert document_1.update(set={"name": "beer"}).name == "beer"


def test_two_fields(document_1: SimpleDocument):
    item: SimpleDocument = document_1.update(
        set={"name": "beer", "price": 2.5}
    )
    assert item.name == "beer"
    assert item.price == 2.5


def test_nested(nested_document_1: NestedDocument):
    new: NestedDocument = nested_document_1.update(
        set={
            "nested.key1.key2": 2
        }
    )

    assert new.nested["key1"]["key2"] == 2


def test_nested_unexistent(nested_document_1: NestedDocument):
    new: NestedDocument = nested_document_1.update(
        set={
            "nested.key1.key3": 4
        }
    )

    assert new.nested["key1"]["key3"] == 4



import contextlib
import os
from pathlib import Path
from types import NoneType

from antievil import NotFoundError

from orwynn.app import AppMode
from orwynn.apprc.apprc import AppRc
from orwynn.utils import validation
from orwynn.utils.klass import Static
from orwynn.utils.mp import patch as mp_patch
from orwynn.utils.yml import load_yml

from .constants import APP_RC_MODE_NESTING


class AppRCUtils(Static):
    @staticmethod
    def parse(
        root_dir: Path,
        mode: AppMode,
        direct_apprc: AppRc | None
    ) -> AppRc:
        """Parses apprc dictionary.

        Args:
            root_dir:
                Root dir of app.
            mode:
                Boot mode of app.
            direct_apprc:
                AppRc raw configuration passed directly instead of path to it
                via environ. This arg is not optional and should accept None
                directly, if such apprc is not present to enable environ
                searching way.
        """
        validation.validate(root_dir, Path)
        validation.validate(mode, AppMode)
        validation.validate(direct_apprc, [AppRc, NoneType])

        # All required for this enabled mode data goes here
        final_apprc: AppRc = {}

        if not direct_apprc:
            rc_paths: list[Path] = [
                Path(root_dir, "orwynn.yml"),
            ]

            env_path: str | None = os.getenv("ORWYNN_RC_PATH", None)
            if env_path is not None:
                final_path: Path = Path(root_dir, env_path)
                if not final_path.exists():
                    raise NotFoundError(
                        title="specified in environ path",
                        value=final_path
                    )
                # look only for environ path
                rc_paths = [final_path]

            for p in rc_paths:
                if p.exists():
                    # Here goes all data contained in yaml config
                    apprc: AppRc = load_yml(p)
                    AppRCUtils._parse_into(
                        final_apprc, apprc, False, mode
                    )
                    break
        else:
            # For direct app rc emptiness check should be always performed, so
            # we pass True
            AppRCUtils._parse_into(final_apprc, direct_apprc, True, mode)

        return final_apprc

    @staticmethod
    def _parse_into(
        receiver: dict,
        source: dict,
        should_check_if_source_empty: bool,
        mode: AppMode
    ) -> None:
        if source == {} and should_check_if_source_empty:
            raise ValueError("parsed apprc source is empty")

        # Check if apprc contains any unsupported top-level keys
        for k in receiver:
            supported_top_level_keys: list[str] = [
                x.value for x in AppMode
            ]
            if k not in supported_top_level_keys:
                raise ValueError(
                    f"unsupported top-level key \"{k}\" of apprc config"
                )

        # Load from bottom to top updating previous one with newest one
        mode_nesting_index: int = APP_RC_MODE_NESTING.index(mode)
        for nesting_mode in APP_RC_MODE_NESTING[:mode_nesting_index + 1]:
            # Supress: We don't mind if any top-level key is missing here
            with contextlib.suppress(KeyError):
                mp_patch(
                    receiver,
                    source[nesting_mode.value],
                    should_deepcopy=False
                )



import io
from contextlib import redirect_stdout
from typing import Callable


def capture_stdout(fn: Callable, *args, **kwargs) -> str:
    """Take stdout of given function call."""
    stringio = io.StringIO()
    with redirect_stdout(stringio):
        fn(*args, **kwargs)
    return stringio.getvalue()



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from orwynn.mapping import Mapping


class MappingIdAlreadySetError(Exception):
    pass


class CustomUseOfMappingReservedFieldError(Exception):
    pass


class UnsetIdMappingError(Exception):
    """
    Id for document has not been set yet.
    """
    def __init__(
        self,
        *,
        explanation: str,
        mapping: "Mapping"
    ) -> None:
        message: str = \
            f"{explanation}: id for mapping <{mapping}> has not been set yet"
        super().__init__(message)


class MappingNotLinkedError(Exception):
    """Is mapping is not linked to accoridng object in database."""



from orwynn.utils.condition import ComparisonCondition, ComparisonMark
from orwynn.utils.expectation import ListExpectation

one_item_list_expectation: ListExpectation = ListExpectation(
    count=ComparisonCondition(
        mark=ComparisonMark.Equal,
        value=1,
    ),
)



from typing import Any, Callable

from fastapi import params
from fastapi.dependencies.models import Dependant
from fastapi.dependencies.utils import (
    add_non_field_param_to_dependency,
    add_param_to_fields,
    get_param_field,
    get_param_sub_dependant,
    get_typed_signature,
    is_scalar_field,
    is_scalar_sequence_field,
)
from fastapi.utils import get_path_param_names


def get_dependant(
    *,
    path: str,
    call: Callable[..., Any],
    name: str | None = None,
    security_scopes: list[str] | None = None,
    use_cache: bool = True,
) -> Dependant:
    path_param_names = get_path_param_names(path)
    endpoint_signature = get_typed_signature(call)
    signature_params = endpoint_signature.parameters
    dependant = Dependant(
        call=call,
        name=name,
        path=path,
        security_scopes=security_scopes,
        use_cache=use_cache,
    )
    for param_name, param in signature_params.items():
        # Ignore orwynn's framework internal arguments
        if param_name.startswith("_fw_"):
            continue

        if isinstance(param.default, params.Depends):
            sub_dependant = get_param_sub_dependant(
                param=param, path=path, security_scopes=security_scopes
            )
            dependant.dependencies.append(sub_dependant)
            continue
        if add_non_field_param_to_dependency(param=param, dependant=dependant):
            continue
        param_field = get_param_field(
            param=param, default_field_info=params.Query, param_name=param_name
        )
        if param_name in path_param_names:
            if is_scalar_field(field=param_field):
                raise TypeError(
                    "Path params must be of one of the supported types"
                )
            ignore_default = not isinstance(param.default, params.Path)
            param_field = get_param_field(
                param=param,
                param_name=param_name,
                default_field_info=params.Path,
                force_type=params.ParamTypes.path,
                ignore_default=ignore_default,
            )
            add_param_to_fields(field=param_field, dependant=dependant)
        elif is_scalar_field(field=param_field):
            add_param_to_fields(field=param_field, dependant=dependant)
        elif isinstance(
            param.default, (params.Query, params.Header)
        ) and is_scalar_sequence_field(param_field):
            add_param_to_fields(field=param_field, dependant=dependant)
        else:
            field_info = param_field.field_info
            if isinstance(field_info, params.Body):
                raise TypeError(
                    f"Param: {param_field.name} can only be a"
                    " request body, using Body()"
                )
            dependant.body_params.append(param_field)
    return dependant



from starlette.exceptions import HTTPException as StarletteHTTPException

HttpException = StarletteHTTPException


class UnsupportedHttpMethodError(Exception):
    pass


class EndpointNotFoundError(Exception):
    pass



from orwynn.utils.condition._mark import ComparisonMark


class UnsupportedComparisonError(Exception):
    """
    Objects of type does not support certain comparison.
    """
    def __init__(
        self,
        *,
        Type: type,
        compare_mark: ComparisonMark,
    ) -> None:
        message: str = \
            f"objects of type <{Type}> does not support" \
            f" comparison <{compare_mark}>"
        super().__init__(message)



# A request of any supported protocol
from enum import Enum
from typing import TYPE_CHECKING, Union

from orwynn.utils.scheme import Scheme

if TYPE_CHECKING:
    from orwynn.http import HttpRequest, HttpResponse
    from orwynn.websocket import Websocket


GenericRequest = Union[
    "HttpRequest",
    # The Websocket object is a representation of the request
    "Websocket"
]
GenericResponse = Union[
    "HttpResponse",
    # The Websocket protocol doesn't return any response in direct form from a
    # controller - it sends it over special channel instead.
    None
]


class RequestMethod(Enum):
    """
    All possible request methods.
    """
    GET = "get"
    POST = "post"
    PUT = "put"
    DELETE = "delete"
    PATCH = "patch"
    OPTIONS = "options"
    WEBSOCKET = "websocket"


REQUEST_METHOD_BY_PROTOCOL: dict[Scheme, list[RequestMethod]] = {
    Scheme.HTTP: [
        RequestMethod.GET,
        RequestMethod.POST,
        RequestMethod.PUT,
        RequestMethod.DELETE,
        RequestMethod.PATCH,
        RequestMethod.OPTIONS,
    ],
    Scheme.WEBSOCKET: [
        RequestMethod.WEBSOCKET
    ]
}



from typing import Any, Callable

from pymongo import MongoClient, ReturnDocument
from pymongo.client_session import ClientSession
from pymongo.cursor import Cursor
from pymongo.database import Database as PymongoDatabase

from orwynn.base.database import Database
from orwynn.base.database.errors import DatabaseEntityNotFoundError
from orwynn.mongo.config import MongoConfig
from orwynn.mongo.entity import MongoEntity
from orwynn.utils import validation


class Mongo(Database):
    """Manages actions related to MongoDB."""
    def __init__(self, config: MongoConfig) -> None:
        self.__client: MongoClient = MongoClient(config.url)
        self.__database: PymongoDatabase = self.__client[config.database_name]
        self.start_session: Callable[[], ClientSession] = \
            self.__client.start_session

    def drop_database(self) -> None:
        self.__client.drop_database(self.__database)

    def find_all(
        self,
        collection: str,
        query: dict,
        **kwargs
    ) -> Cursor:
        validation.validate(collection, str)
        validation.validate(query, dict)

        return self.__database[collection].find(
            query, **kwargs
        )

    def find_one(
        self, collection: str, query: dict, **kwargs
    ) -> MongoEntity:
        validation.validate(collection, str)
        validation.validate(query, dict)

        result: Any | None = self.__database[collection].find_one(
            query, **kwargs
        )
        if result is None:
            raise DatabaseEntityNotFoundError(
                query=query, collection=collection
            )
        return validation.apply(result, dict)

    def create_one(
        self, collection: str, document: MongoEntity, **kwargs
    ) -> MongoEntity:
        """Creates a document returning it after creation."""
        validation.validate(collection, str)
        validation.validate(document, dict)

        inserted_id: str = self.__database[collection].insert_one(
            document, **kwargs
        ).inserted_id
        copied: MongoEntity = document.copy()

        # instead of searching for created document, just replace it's id
        # with mongo's generated one, which is better for performance
        copied["_id"] = inserted_id

        return copied

    def remove_one(
        self, collection: str, query: dict, **kwargs
    ) -> MongoEntity:
        """Deletes a document matching query and returns it."""
        validation.validate(collection, str)
        validation.validate(query, dict)

        removed_document: Any | None = \
            self.__database[collection].find_one_and_delete(
                query, **kwargs
            )

        if removed_document is None:
            raise DatabaseEntityNotFoundError(
                collection=collection, query=query
            )

        return validation.apply(removed_document, MongoEntity)

    def update_one(
        self,
        collection: str,
        query: dict,
        operation: dict,
        **kwargs
    ) -> MongoEntity:
        """Updates a document matching query and returns updated version."""
        validation.validate(collection, str)
        validation.validate(query, dict)
        validation.validate(operation, dict)

        updated_document: Any = \
            self.__database[collection].find_one_and_update(
                query,
                operation,
                return_document=ReturnDocument.AFTER,
                **kwargs
            )

        if updated_document is None:
            raise DatabaseEntityNotFoundError(
                collection=collection, query=query
            )

        return validation.apply(updated_document, MongoEntity)



from typing import TYPE_CHECKING, TypeVar

if TYPE_CHECKING:
    from orwynn.mongo import Document

MongoCompatibleType = str | int | float | bool | list | dict | None
TDocument = TypeVar("TDocument", bound="Document")



import os
from collections.abc import Callable
from typing import TYPE_CHECKING

from antievil import NotFoundError
from sqlalchemy import Select, text

from orwynn.base.model import Model
from orwynn.di.di import Di
from orwynn.log import Log
from orwynn.sql.search import TableSearch
from orwynn.sql.shd import SHD
from orwynn.sql.sql import SQL
from orwynn.sql.stateflag import StateFlag
from orwynn.sql.table import Table
from orwynn.sql.types import ConvertedModel, ListedConvertedModel, TTable
from orwynn.utils import validation
from orwynn.utils.klass import Static

if TYPE_CHECKING:
    from collections.abc import Sequence


class SHDUtils(Static):
    @classmethod
    def scalars_all(
        cls,
        selection: Select[tuple[TTable]],
        search: TableSearch[TTable],
        shd: SHD | None = None,
    ) -> list[TTable]:
        """
        Executes a selection statement and returns result as all scalars.

        Args:
            selection:
                Selection to execute.
            search:
                What search is performed to produce given selection statement.
            shd(optional):
                Explicitly given SHD in case if search was performed using
                internal SHD. Defaults to search's SHD, i.e. external one.
        """
        final_shd: SHD = shd if shd is not None else search.get_shd()

        result: Sequence[TTable] | None = final_shd.scalars(
            selection,
        ).all()

        if not result:
            # we just give "T" here as a title since i don't want to
            # investigate generic's name retrieval in Python -- ryzhovalex
            raise search.get_not_found_error("TTable")
        if search.expectation is not None:
            search.expectation.check(result)

        return list(result)


class SQLUtils(Static):
    @staticmethod
    def drop_database(
        sql: SQL,
    ) -> None:
        if (
            os.environ.get("HQB_SHOULD_DROP_SQL", "0") == "1"
            # TODO(ryzhovalex): temporarily disabled PROD checking
            # 0
            #
            # and BootProxy.ie().mode != AppMode.PROD
        ):
            Log.info("drop sql database")
            # use raw sql since i haven't found a quick way to drop cascade
            # using python objects
            with sql.session as s:
                s.execute(text("""
                DROP SCHEMA public CASCADE;
                CREATE SCHEMA public;
                """))
                s.commit()

    @staticmethod
    def get_state_flag_by_name(
        name: str,
        shd: SHD,
    ) -> StateFlag:
        validation.validate(name, str)

        result: StateFlag | None = shd.scalar(
            shd.select(StateFlag).where(
                StateFlag.name == name,
            ),
        )

        if not result:
            raise NotFoundError(
                title="state flag",
                options={
                    "name": name,
                },
            )

        return validation.apply(
            result,
            StateFlag,
        )

    @staticmethod
    def set_state_flag_by_name(
        name: str,
        value: bool,
        shd: SHD,
    ) -> StateFlag:
        """
        Finds a state flag by name and sets a new value to it.
        """
        validation.validate(name, str)
        validation.validate(value, bool)

        state_flag: StateFlag = SQLUtils.get_state_flag_by_name(name, shd)
        state_flag.value = value

        return state_flag

    @staticmethod
    def get_one(
        id: str,
        Table_: type[TTable],
        shd: SHD,
    ) -> TTable:
        """Finds table row with given id.

        Works similar as sqlalchemy.shd.get(), but with id.

        Args:
            id:
                Id to match.
            Table_:
                Table to search within.
            session:
                SqlA shd.

        Returns:
            Table row found.
        """
        validation.validate(id, str)
        validation.validate(Table_, Table)
        validation.validate(shd, SHD)

        result: TTable | None = shd.scalar(
            shd.select(Table_).where(
                Table_.id == id,
            ),
        )

        if not result:
            raise NotFoundError(
                title="table",
                options={
                    "id": id,
                },
            )

        return validation.apply(
            result,
            Table_,
        )

    @staticmethod
    def get_all_by_ids(
        ids: list[str],
        Table_: type[TTable],
        shd: SHD,
    ) -> list[TTable]:
        """Finds all table rows using list of ids.

        Args:
            ids:
                Ids to find.
            Table_:
                Table to search within.
            session:
                SqlA shd.

        Returns:
            Table rows found.
        """
        validation.validate(ids, list)

        rows: list[TTable] = []

        for id in ids:
            rows.append(
                SQLUtils.get_one(id, Table_, shd),
            )

        return rows

    @staticmethod
    def get_one_and_convert(
        id: str,
        Table_: type[TTable],
        convertion_fn: Callable[[TTable], ConvertedModel],
        shd: SHD,
    ) -> ConvertedModel:
        """
        Find one row and converts it into model using convertion function.

        Args:
            id:
                Id to match.
            Table_:
                Table where object resides.
            convertion_fn:
                Convertion function to be used to convert Table->Model.
            session:
                SqlA shd.

        Returns:
            Converted model.
        """
        validation.validate(Table_, Table)
        validation.validate(convertion_fn, Callable)
        validation.validate(shd, SHD)

        return convertion_fn(SQLUtils.get_one(
            id,
            Table_,
            shd,
        ))

    @staticmethod
    def get_all_and_convert(
        Table_: type[TTable],
        convertion_fn: Callable[[TTable], ConvertedModel],
        ListedConvertedModel_: type[ListedConvertedModel],
        listed_converted_model_units_key: str,
        shd: SHD,
    ) -> ListedConvertedModel:
        """
        Finds all rows for table and converts them into listed model (model
        that contains many converted models).

        Args:
            Table_:
                Table where object resides.
            convertion_fn:
                Convertion function to be used to convert Table->Model.
            ListedConvertedModel_:
                Model to be used as a container for all converted models.
            listed_converted_model_units_key:
                Key of the ListedConvertedModel_ where collected units should
                be passed. For example "products".
            session:
                SqlA shd.

        Returns:
            Listed model with converted models.
        """
        validation.validate(Table_, Table)
        validation.validate(convertion_fn, Callable)
        validation.validate(ListedConvertedModel_, Model)
        validation.validate(listed_converted_model_units_key, str)
        validation.validate(shd, SHD)

        rows: list[TTable] = SQLUtils.get_all(
            Table_,
            shd,
        )
        converted_list: list[ConvertedModel] = []
        for row in rows:
            converted_list.append(convertion_fn(row))
        return ListedConvertedModel_.parse_obj({
            listed_converted_model_units_key: converted_list,
        })

    @staticmethod
    def get_all(
        Table_: type[TTable],
        shd: SHD,
        *,
        ids: list[str] | None = None,
    ) -> list[TTable]:
        """
        Find all rows for table.

        Args:
            Table_:
                Table to search within.
            session:
                SqlA shd.
            ids (optional):
                List of ids for which to found tables.

        Returns:
            All rows for table.
        """
        validation.validate(shd, SHD)
        validation.validate(Table_, Table)

        selection = shd.select(Table_)

        if ids:
            selection = selection.where(
                Table_.id.in_(ids),
            )

        result: Sequence[TTable] | None = shd.scalars(
            selection,
        ).all()

        if not result:
            raise NotFoundError(
                title="table",
                options={
                    "ids": ids,
                },
            )

        tables: list[TTable] = list(result)
        validation.validate_each(tables, Table)

        return tables

    @staticmethod
    def get_sql_from_di(di: Di) -> SQL:
        return validation.apply(di.find("SQL"), SQL)

    @staticmethod
    def create_tables(sql: SQL):
        sql.create_tables()



from orwynn.base.config import Config
from orwynn.base.controller import Controller
from orwynn.base.database import Database
from orwynn.base.errorhandler import ErrorHandler
from orwynn.base.middleware import Middleware
from orwynn.base.model import Model
from orwynn.base.module import Module
from orwynn.base.service import Service
from orwynn.base.worker import Worker



from antievil import WrongGenericTypeError

from orwynn.utils import validation
from orwynn.utils.condition import ComparisonCondition
from orwynn.utils.condition._mark import ComparisonMark
from orwynn.utils.condition.errors import UnsupportedComparisonError


def test_int_equal():
    condition: ComparisonCondition[int] = ComparisonCondition(
        ComparisonMark.Equal,
        value=5,
    )

    assert condition.compare(5)
    assert not condition.compare(10)


def test_int_wrong_type():
    condition: ComparisonCondition[int] = ComparisonCondition(
        ComparisonMark.Equal,
        value=5,
    )

    validation.expect(
        condition.compare,
        WrongGenericTypeError,
        "impostor",
    )


def test_int_less_equal():
    condition: ComparisonCondition[int] = ComparisonCondition(
        ComparisonMark.LessEqual,
        value=5,
    )

    assert condition.compare(1)
    assert condition.compare(5)
    assert not condition.compare(10)


def test_unsupported_comparison():
    condition: ComparisonCondition[dict[str, int]] = ComparisonCondition(
        ComparisonMark.LessEqual,
        value={"hello": 1},
    )

    validation.expect(
        condition.compare,
        UnsupportedComparisonError,
        {
            "world": 2,
        },
    )



from typing import Callable

import pytest

from orwynn.base.module.module import Module
from orwynn.boot.boot import Boot
from orwynn.http import (
    Endpoint,
    HttpController,
    HttpMiddleware,
    HttpRequest,
    HttpResponse,
    TestHttpResponse,
)
from orwynn.testing.client import Client


class Mw1(HttpMiddleware):
    async def process(
        self, request: HttpRequest, call_next: Callable
    ) -> HttpResponse:
        response: HttpResponse = await call_next(request)
        response.headers["x-test"] = "hello"
        return response


@pytest.mark.asyncio
async def test_basic():
    """
    Should work in general cases.
    """
    class C1(HttpController):
        Route = "/"
        Endpoints = [Endpoint(method="get")]

        def get(self):
            return {"message": "hello"}

    boot: Boot = await Boot.create(Module(
        route="/hello/world",
        Controllers=[C1],
        Middleware=[Mw1]
    ))
    http: Client = boot.app.client
    response: TestHttpResponse = http.get("/hello/world")

    assert response.headers["x-test"] == "hello"


@pytest.mark.asyncio
async def test_variable_route():
    """Should work with variable routes.
    """
    class C1(HttpController):
        Route = "/{id}"
        Endpoints = [Endpoint(method="get")]

        def get(self, id: str):
            return {"value": id}

    boot: Boot = await Boot.create(Module(
        route="/",
        Controllers=[C1],
        Middleware=[Mw1]
    ))
    http: Client = boot.app.client
    response: TestHttpResponse = http.get("/e201")

    assert response.json()["value"] == "e201"
    assert response.headers["x-test"] == "hello"


@pytest.mark.asyncio
async def test_file_path_route():
    """Should be OK with file paths.
    """
    class C1(HttpController):
        Route = "/{file_path:path}"
        Endpoints = [Endpoint(method="get")]

        def get(self, file_path: str):
            return {"value": file_path}

    boot: Boot = await Boot.create(Module(
        route="/",
        Controllers=[C1],
        Middleware=[Mw1]
    ))
    http: Client = boot.app.client
    response: TestHttpResponse = http.get("/doc/pdf/1234.pdf")

    assert response.json()["value"] == "doc/pdf/1234.pdf"
    assert response.headers["x-test"] == "hello"



from typing import Generic, Literal, Self

from antievil import NameExpectError, UnsupportedError

from orwynn.sql.shd import SHD
from orwynn.sql.table import Table
from orwynn.sql.types import TTable
from orwynn.utils.dbsearch import DatabaseSearch


class TableSearch(DatabaseSearch[Table], Generic[TTable]):
    """
    Base search terms to find SQL tables.

    Used as a main argument to most Repo table getters.
    """
    shd: SHD | Literal["internal"]
    """
    Session handler of the search.

    Can be set to literal "internal" signifying that a search is meant to work
    with callable's internal SHD. This is a typical case when working e.g. with
    DTO's repos, which methods work only with internal shds in order to bake
    correct dto models on return.
    """

    class Config:
        arbitrary_types_allowed = True

    def get_shd(self) -> SHD:
        if isinstance(self.shd, SHD):
            return self.shd
        raise UnsupportedError(
            title="for external using, search shd",
            value=self.shd,
        )

    def check_internal(self) -> None:
        if self.shd != "internal":
            raise NameExpectError(
                objinfo=("shd", self.shd),
                name="internal",
            )

    def activate_shd(self, shd: SHD) -> Self:
        """
        Return copy of self with SHD's activated.

        Only works for search models with shd="internal".
        """
        if self.shd != "internal":
            raise UnsupportedError(
                title="activating shd for search with existing shd",
                value=shd,
            )
        return self.copy(update={"shd": shd})



from enum import Enum


class CreateOrUpdateStatus(Enum):
    """
    When instance could have been created or updated depending of whether it
    have been found or not.
    """
    Created = "created"
    Updated = "updated"


class SQLDatabaseKind(Enum):
    PostgreSQL = "postgresql"
    SQLite = "sqlite"



from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from orwynn.utils.expectation import Expectation


class ExpectationError(Exception):
    """
    Some expectation rule is failed.
    """
    def __init__(
        self,
        *,
        title: str,
        actual_value: Any,
        expected_value: Any,
    ) -> None:
        message: str = \
            f"{title} expected to have value <{expected_value}>," \
            f" got <{actual_value}> instead"
        super().__init__(message)


class UnsupportedExpectationTypeError(Exception):
    """
    Type is not supported by the expectation.
    """
    def __init__(
        self,
        *,
        UnsupportedType: type,
        expectation: "Expectation",
    ) -> None:
        message: str = \
            f"expectation <{expectation}> does not support" \
            f" type <{UnsupportedType}>"
        super().__init__(message)



from typing import Any


class RestrictedDeletionError(Exception):
    """
    When deletion of object is restricted due to dependency, e.g. in
    one-to-many relations: you cannot delete parent if it has any children in
    relationship - they would end up orphaned which is often restricted by
    Sql NOT NULL constraint for foreign key.
    """

class AlreadyAttachedError(Exception):
    """
    When someone tries to attach an object in relationship that already exist
    there.
    """
    def __init__(
        self,
        *,
        attachment_name: str,
        attachment_value: str,
        target_name: str,
        target_value: str,
    ) -> None:
        message: str = \
            f"{attachment_name}={attachment_value} already attached" \
            f" to {target_name}={target_value}"
        super().__init__(message)


class NonManageableSHDError(Exception):
    """
    Non-manageable SHD cannot perform some operation.
    """
    def __init__(
        self,
        *,
        operation_name: str,
    ):
        message: str = \
            f"cannot perform operation={operation_name} in a non-manageable" \
            " session"
        super().__init__(message)


class ManageableSHDError(Exception):
    """
    Manageable SHD cannot perform some operation.
    """
    def __init__(
        self,
        *,
        operation_name: str,
    ):
        message: str = \
            f"cannot perform operation={operation_name} in a manageable" \
            " session"
        super().__init__(message)


class EmptyExecutionQueueError(Exception):
    """
    Execution queue is empty.
    """
    def __init__(
        self,
    ):
        message: str = "execution queue is empty"
        super().__init__(message)


class SqlToUpstreamUnmatchError(Exception):
    """
    SQL cannot be None if an upstream SHD is not set.
    """
    def __init__(
        self,
    ):
        message: str = \
            "cannot set SQL to None since the upstream shd is None too"
        super().__init__(message)


class UnexpectedAmountError(Exception):
    """
    Not expected amount of instances are returned.
    """
    def __init__(
        self,
        *,
        title: str,
        expected_amount: int,
        fact_amount: int,
    ):
        message: str = \
            f"{title} amount expected={expected_amount}," \
            f" got fact={fact_amount}"
        super().__init__(message)


class ExistingUnitError(Exception):
    """
    Unit already exists in database.
    """
    def __init__(
        self,
        *,
        title: str,
        value: Any,
    ):
        message: str = f"{title}={value} already exists in database"
        super().__init__(message)



import typing
from inspect import isclass
from typing import Any, Callable, Literal

import pydantic

from orwynn.apiversion import ApiVersion
from orwynn.app import App
from orwynn.base.controller.controller import Controller
from orwynn.base.controller.errors import AlreadyRegisteredMethodError
from orwynn.base.error import MalfunctionError
from orwynn.base.model import Model
from orwynn.base.module import Module
from orwynn.helpers.web import REQUEST_METHOD_BY_PROTOCOL, RequestMethod
from orwynn.http import Endpoint, EndpointContainer, HttpController
from orwynn.http.errors import (
    EndpointNotFoundError,
    UnsupportedHttpMethodError,
)
from orwynn.indication.indication import Indication
from orwynn.proxy.boot import BootProxy
from orwynn.router.errors import (
    UnmatchedEndpointEntityError,
)
from orwynn.utils import validation
from orwynn.utils.scheme import Scheme
from orwynn.utils.url import join_routes
from orwynn.websocket import (
    WebsocketController,
    WebsocketStack,
    routing_handlers,
)


class ControllerRegister:
    """
    Registers controllers to the system.
    """
    def __init__(
        self,
        *,
        app: App,
        modules: list[Module],
        controllers: list[Controller],
        websocket_stack: WebsocketStack,
        global_http_route: str,
        global_websocket_route: str,
        api_version: ApiVersion
    ) -> None:
        self.__app: App = app

        self.__modules: list[Module] = modules
        self.__controllers: list[Controller] = controllers
        self.__methods_by_route: dict[str, set[RequestMethod]] = {}

        self.__global_http_route: str = global_http_route
        self.__global_websocket_route: str = global_websocket_route
        self.__api_version: ApiVersion = api_version

        self.__websocket_stack: WebsocketStack = websocket_stack

    def register_all(
        self
    ) -> None:
        """
        Registers all controllers to the system.
        """
        for m in self.__modules:
            for C in m.Controllers:
                self.__register_controller_class_for_module(
                    m, C
                )

    def __register_controller_class_for_module(
        self,
        m: Module,
        C: type[Controller]
    ) -> None:
        if m.route is None:
            raise MalfunctionError(
                f"module {m} has not route and shouldn't have added any"
                " controllers"
            )

        is_controller_found: bool = False
        for c in self.__controllers:
            if type(c) is C:
                is_controller_found = True

                if isinstance(c, HttpController):
                    self.__register_http_controller_for_module(c, m)
                elif isinstance(c, WebsocketController):
                    # Websocket controllers are firstly added and only then
                    # registered to form a correct stack
                    self.__add_websocket_controller_for_module(
                        c,
                        m
                    )
                else:
                    raise TypeError(
                        f"controller unsupported type {type(c)}"
                    )
        if not is_controller_found:
            raise MalfunctionError(
                f"no initialized controller found for class {C},"
                f" but it was declared in imported module {m},"
                " so DI should have been initialized it"
            )

    def __add_websocket_controller_for_module(
        self, controller: WebsocketController, module: Module
    ) -> None:
        """Add websocket controller for a route.

        Args:
            controller:
                Websocket controller to add.
            module:
                Module to register for.
        """
        # Methods started from "on_" or equal to "main" should be registered.
        # "main" is assigned to MODULE_Route + CONTROLLER_Route directly.
        for event_handler in controller.event_handlers:
            method_route: str = \
                "/" if event_handler.name == "main" else event_handler.name
            routes: set[str] = self.__get_routes_for_controller(
                controller,
                module,
                additional_route=method_route
            )

            controller._fw_update_final_routes(routes)

            for route in routes:
                self.__websocket_stack.add_call(
                    routing_handlers.WebsocketHandler(
                        fn=event_handler.fn,
                        route=route
                    )
                )

    def __register_http_controller_for_module(
        self,
        controller: HttpController,
        module: Module
    ) -> None:
        routes: set[str] = self.__get_routes_for_controller(
            controller,
            module
        )

        # At least one method found
        is_method_found: bool = False
        for http_method in RequestMethod:
            # Don't register unused methods
            if (
                http_method in controller.methods
                and http_method in REQUEST_METHOD_BY_PROTOCOL[Scheme.HTTP]
            ):
                is_method_found = True

                for route in routes:
                    self.__register_http_route(
                        route=route,
                        fn=controller.get_fn_by_http_method(http_method),
                        method=http_method
                    )

        if not is_method_found:
            raise MalfunctionError(
                f"no http methods found for controller {controller.__class__},"
                " this shouldn't have passed validation at Controller.__init__"
            )
        else:
            # Make the controller know about it's full routes
            controller._fw_update_final_routes(routes)

    def __get_routes_for_controller(
        self,
        controller: Controller,
        module: Module,
        *,
        additional_route: str = ""
    ) -> set[str]:
        """
        Returns all routes which given controller accessible from.

        Args:
            ...
            additional_route (optional):
                Route to append addiotionally to the end of normally
                concatenated route. Defaults to empty string.
        """
        if module.route is None:
            raise MalfunctionError(
                f"module {module} has not route and shouldn't have added any"
                " controllers"
            )

        routes: set[str] = set()
        final_versions: set[int] = set()
        version: int | set[int] | None | Literal["*"] = controller.Version

        # Get the latest version if none is specified for controller
        if version is None:
            final_versions.add(self.__api_version.latest)
        elif isinstance(version, int):
            final_versions.add(version)
        elif isinstance(version, set):
            final_versions.update(version)
        elif isinstance(version, str):
            if version != "*":
                raise MalfunctionError(
                    f"version cannot be {version}, a validation check should"
                    " have been performed at HTTPController"
                )
            # Add all supported versions
            final_versions.update(self.__api_version.supported)
        else:
            raise MalfunctionError(
                f"unrecognized version {version}, a validation check should"
                " have been performed at HTTPController"
            )

        for v in final_versions:
            final_global_route: str = self.__get_global_route_for_controller(
                controller,
                api_version_number=v
            )

            concatenated_route: str
            if isinstance(controller, HttpController):
                # We can concatenate routes such way since routes
                # are validated to not contain following slash
                # -> But join_routes() handles this situation, doesn't it?
                concatenated_route = join_routes(
                    final_global_route,
                    module.route,
                    controller.route,
                    additional_route
                )
            elif isinstance(controller, WebsocketController):
                concatenated_route = join_routes(
                    final_global_route,
                    module.route,
                    controller.route,
                    additional_route
                )
            else:
                raise TypeError(
                    f"unrecognized controller {controller}"
                )
            routes.add(concatenated_route)

        return routes

    def __get_global_route_for_controller(
        self,
        controller: Controller,
        *,
        api_version_number: int
    ) -> str:
        # Choose which protocol's global route to use
        global_route: str
        if isinstance(controller, HttpController):
            global_route = self.__global_http_route
        elif isinstance(controller, WebsocketController):
            global_route = self.__global_websocket_route
        else:
            raise TypeError(
                f"unrecognized controller {controller}"
            )

        try:
            return self.__api_version.apply_version_to_route(
                global_route,
                api_version_number
            )
        except ValueError:
            # No {version} format block
            return global_route

    def __parse_endpoint_spec_kwargs(
        self, spec: Endpoint | None, fn: Callable
    ) -> dict[str, Any]:
        result: dict[str, Any] = {}
        fn_return_typehint: Any | None = typing.get_type_hints(fn).get(
            "return", None
        )

        if spec is not None:
            result["status_code"] = spec.default_status_code
            result["summary"] = spec.summary
            result["tags"] = spec.tags
            result["deprecated"] = spec.is_deprecated

            api_indication: Indication = BootProxy.ie().api_indication
            final_responses: dict[int, dict[str, Any]] = {}
            if spec.responses:
                for response in spec.responses:
                    if (
                        response.status_code == spec.default_status_code
                        and fn_return_typehint is not dict
                        and response.Entity is not fn_return_typehint
                    ):
                        raise UnmatchedEndpointEntityError(
                            f"route handler {fn} response endpoint entity"
                            f" {response.Entity} is not match returned"
                            f" typehint {fn_return_typehint}"
                        )

                    final_responses[response.status_code] = {
                        "model": api_indication.gen_schema(response.Entity),
                        "description": response.description
                    }

            # Add default status code response AKA response_model
            if spec.default_status_code not in final_responses:
                if (
                    fn_return_typehint is not None
                    and isclass(fn_return_typehint)
                    and issubclass(fn_return_typehint, Model)
                ):
                    final_responses[spec.default_status_code] = {
                        "model": api_indication.gen_schema(
                            fn_return_typehint
                        )
                    }
                else:
                    result["response_model"] = None

            # Add default pydantic validation error
            if (
                422 not in final_responses
                and self.__is_pydantic_validation_error_can_occur_in_fn(fn)
            ):
                final_responses[422] = {
                    "model": api_indication.gen_schema(
                        pydantic.ValidationError
                    ),
                    "description": "Validation Error"
                }
            result["responses"] = final_responses

        return result

    def __is_pydantic_validation_error_can_occur_in_fn(
        self,
        fn: Callable
    ) -> bool:
        # Wrap return statement in try...except to prevent error from
        # issubclass() about wrong typehint (typehint is not a class)
        for typehint in typing.get_type_hints(fn).values():
            try:
                flag: bool = issubclass(typehint, pydantic.BaseModel)
            except TypeError:
                continue
            else:
                if flag:
                    return True
        return False

    def __register_http_route(
        self, *, route: str, fn: Callable, method: RequestMethod
    ) -> None:
        """Registers a fn for a route.

        Attributes:
            route:
                Route to register to.
            fn:
                Function to register.
            method:
                HTTP method function is handling.
        """
        validation.validate(route, str)
        validation.validate(fn, Callable)
        validation.validate(method, RequestMethod)

        app_fn: Callable | None = \
            self.__app.HTTP_METHODS_TO_REGISTERING_FUNCTIONS.get(
                method, None
            )

        if not app_fn:
            raise UnsupportedHttpMethodError(
                f"HTTP method {method} is not supported"
            )

        if (
            route in self.__methods_by_route
            and method in self.__methods_by_route[route]
        ):
            raise AlreadyRegisteredMethodError(
                f"method {method} has been already registered for route"
                f" \"{route}\""
            )

        try:
            self.__methods_by_route[route].add(method)
        except KeyError:
            self.__methods_by_route[route] = {method}

        spec: Endpoint | None
        try:
            spec = EndpointContainer.ie().find_spec(fn)
        except EndpointNotFoundError:
            spec = None

        app_fn(
            route,
            **self.__parse_endpoint_spec_kwargs(
                spec,
                fn
            )
        )(fn)



from orwynn.base.config import Config
from orwynn.base.controller.controller import Controller
from orwynn.base.errorhandler.errorhandler import ErrorHandler
from orwynn.base.middleware import Middleware
from orwynn.base.service import FrameworkService, Service
from orwynn.di.acceptor import Acceptor
from orwynn.di.provider import Provider

"""
List of builtin classes in Provider category.

Providers with higher priority cannot inject ones with lower priority.

This list: Lower index => higher priority.
"""
BUILTIN_PROVIDERS: list[type[Provider]] = [
    Config,
    FrameworkService,
    Service
]


"""
List of builtin classes are able to accept Providers.
"""
BUILTIN_ACCEPTORS: list[type[Acceptor]] = [
    # All Providers are Acceptors at the same time
    *BUILTIN_PROVIDERS,
    Controller,
    Middleware,
    ErrorHandler
]




from typing import Any

from orwynn.base.config import Config
from orwynn.base.error import MalfunctionError
from sqlalchemy.pool import StaticPool, Pool

from orwynn.sql.poolclass import PoolclassStr
from .enums import SQLDatabaseKind


class SQLConfig(Config):
    database_kind: SQLDatabaseKind
    database_name: str | None = None
    database_user: str | None = None
    database_password: str | None = None
    database_path: str | None = None
    database_host: str | None = None
    database_port: int | None = None
    poolclass: PoolclassStr | None = None
    pool_size: int | None = None

    def __init__(self, **data: Any) -> None:
        try:
            db_kind: SQLDatabaseKind = SQLDatabaseKind(data["database_kind"])
        except KeyError:
            # memory sqlite is used by default
            db_kind = SQLDatabaseKind.SQLite
            data["database_path"] = ":memory:"

        # Check right associations
        if db_kind is SQLDatabaseKind.PostgreSQL:
            for key in [
                "database_name",
                "database_user",
                "database_password",
                "database_host",
                "database_port"
            ]:
                if not data.get(key, None):
                    raise ValueError(
                        f"for PostgreSQL you should define {key} in SQL config"
                    )
        elif db_kind is SQLDatabaseKind.SQLite:
            if not data.get("database_path", None):
                raise ValueError(
                    "for SQLite you should define database_path in SQL config"
                )
        else:
            raise MalfunctionError()

        super().__init__(**data)

    @property
    def real_poolclass(self) -> type[Pool] | None:
        if self.poolclass is None:
            return None

        match self.poolclass:
            case "StaticPool":
                return StaticPool
            case _:
                raise ValueError()



from orwynn.base.model.model import Model
from orwynn.indication.indicatable import Indicatable


class EndpointResponse(Model):
    status_code: int
    Entity: type[Indicatable]
    description: str | None = None



from orwynn.mongo.document.testing import NestedDocument, SimpleDocument


def test_main(document_1: SimpleDocument):
    item: SimpleDocument = document_1.update(inc={"price": 2.5})
    assert item.price == 3.7


def test_nested(nested_document_1: NestedDocument):
    new: NestedDocument = nested_document_1.update(
        inc={
            "nested.key1.key2": 2
        }
    )

    assert new.nested["key1"]["key2"] == 3


def test_nested_unexistent(nested_document_1: NestedDocument):
    new: NestedDocument = nested_document_1.update(
        # incremention works as set on empty fields
        inc={
            "nested.key1.key3": 3
        }
    )

    assert new.nested["key1"]["key3"] == 3



"""
Works with url formatting.
"""
from orwynn.utils import validation
from orwynn.utils.url.utils import match_routes

from .helpers import get_vars
from .url import Url
from .vars import UrlVars


def join_routes(*routes: str) -> str:
    """Joins all given routes and normalize final result."""
    validation.validate_each(routes, str, expected_sequence_type=tuple)

    result: str = ""

    for route in routes:
        if route == "" or route == "/":
            continue
        elif route[0] != "/":
            result += "/" + route
        else:
            result += route
        result.removesuffix("/")

    if result == "":
        result = "/"

    return result



import pytest

from orwynn.base.module.module import Module
from orwynn.boot.boot import Boot
from orwynn.context.errors import UndefinedStorageError
from orwynn.utils import validation
from orwynn.websocket.context.id import (
    WebsocketRequestContextId,
)
from orwynn.websocket.controller.controller import (
    WebsocketController,
)
from orwynn.websocket.log.middleware import LogWebsocketMiddleware
from orwynn.websocket.websocket import Websocket


@pytest.mark.asyncio
async def test_basic():
    """Request id should be fetchable from context within request-response
    cycle and unfetchable outside this cycle.
    """
    class C1(WebsocketController):
        Route = "/"

        async def main(self, websocket: Websocket) -> None:
            await websocket.send_json(
                {"request_id": WebsocketRequestContextId().get()}
            )

    boot: Boot = await Boot.create(
        Module(
            "/",
            Controllers=[C1],
            Middleware=[LogWebsocketMiddleware]
        )
    )

    validation.expect(
        WebsocketRequestContextId().get,
        UndefinedStorageError
    )

    with boot.app.client.websocket("/") as ws:
        data: dict = ws.receive_json()
        assert type(data["request_id"]) is str and data["request_id"] != ""



from typing import Any

from orwynn.mongo import MongoUtils


def test_s1():
    result: dict[str, Any] = MongoUtils.convert_dict({
        "a1": {
            "a2": "b1",
            "a3": "b2",
        },
    })

    assert result == {
        "a1.a2": "b1",
        "a1.a3": "b2",
    }


def test_s2():
    result: dict[str, Any] = MongoUtils.convert_dict({
        "a1": {
            "a2": {
                "a3": 10,
                "a4": 20,
            },
            "a5": "b1",
            "a6": "b2",
        },
    })

    assert result == {
        "a1.a2.a3": 10,
        "a1.a2.a4": 20,
        "a1.a5": "b1",
        "a1.a6": "b2",
    }


def test_s3():
    result: dict[str, Any] = MongoUtils.convert_dict({
        "a1": {
            "a2": {
                "$in": [1, 2, 3],
            },
            "a3": {
                "$exists": True,
            },
            "a4": {
                "$somevalue": {
                    "notconverted": 1,
                },
            },
            "a5": {
                "somevalue": {
                    "converted": 10,
                },
            },
        },
    })

    assert result == {
        "a1.a2": {
            "$in": [1, 2, 3],
        },
        "a1.a3": {
            "$exists": True,
        },
        "a1.a4": {
            "$somevalue": {
                "notconverted": 1,
            },
        },
        "a1.a5.somevalue.converted": 10,
    }



import pytest

from orwynn.apiversion import ApiVersion
from orwynn.base.module.module import Module
from orwynn.boot.boot import Boot
from orwynn.di.di import Di
from orwynn.utils import validation
from orwynn.websocket import Websocket, WebsocketController


class WsCtrl(WebsocketController):
    Route = "/"
    Version = {2, 3}

    async def main(self, ws: Websocket) -> None:
        await ws.send_json({"message": "hello"})

    async def on_connect(self, ws: Websocket) -> None:
        await ws.send_json({"message": "hello"})


class ArgumentedCtrl(WebsocketController):
    Route = "/user/{user_id}"

    async def main(
        self,
        ws: Websocket,
        user_id: str,
        order: int | None = None,
        message: str | None = "welcome"
    ) -> None:
        await ws.send_json({
            "user_id": user_id,
            "message": message,
            "order": order
        })


@pytest.mark.asyncio
async def test_main_route():
    class WS1(WebsocketController):
        Route = "/hello"

        async def main(self, ws: Websocket) -> None:
            await ws.send_json({"message": "Hello!"})
            await ws.close()

    client = (await Boot.create(Module(
        route="/",
        Controllers=[WS1]
    ))).app.client

    with client.websocket("/hello") as ws:
        data = ws.receive_json()

    assert data == {"message": "Hello!"}


@pytest.mark.asyncio
async def test_custom_route():
    class WS1(WebsocketController):
        Route = "/hello"

        async def on_message(self, ws: Websocket) -> None:
            await ws.send_json({"message": "Hello!"})
            await ws.close()

    client = (await Boot.create(Module(
        route="/",
        Controllers=[WS1]
    ))).app.client

    with client.websocket("/hello/message") as ws:
        data = ws.receive_json()

    assert data == {"message": "Hello!"}


@pytest.mark.asyncio
async def test_several_routes():
    class WS1(WebsocketController):
        Route = "/hello"

        async def main(self, ws: Websocket) -> None:
            await ws.send_json({"message": "main"})
            await ws.close()

        async def on_message(self, ws: Websocket) -> None:
            await ws.send_json({"message": "message"})
            await ws.close()

        async def on_hello(self, ws: Websocket) -> None:
            await ws.send_json({"message": "hello"})
            await ws.close()

    client = (await Boot.create(Module(
        route="/",
        Controllers=[WS1]
    ))).app.client

    with client.websocket("/hello") as ws:
        data = ws.receive_json()
        assert data == {"message": "main"}

    with client.websocket("/hello/message") as ws:
        data = ws.receive_json()
        assert data == {"message": "message"}

    with client.websocket("/hello/hello") as ws:
        data = ws.receive_json()
        assert data == {"message": "hello"}


@pytest.mark.asyncio
async def test_arguments():
    boot: Boot = await Boot.create(
        Module("/", Controllers=[ArgumentedCtrl])
    )

    with boot.client.websocket("/user/eg1?message=hello&order=2") as ws:
        data: dict = ws.receive_json()

        assert data["user_id"] == "eg1"
        assert data["message"] == "hello"
        assert data["order"] == 2


@pytest.mark.asyncio
async def test_default_query():
    boot: Boot = await Boot.create(
        Module("/", Controllers=[ArgumentedCtrl])
    )

    with boot.client.websocket("/user/eg1") as ws:
        data: dict = ws.receive_json()

        assert data["user_id"] == "eg1"
        assert data["message"] == "welcome"
        assert data["order"] is None



@pytest.mark.asyncio
async def test_final_routes():
    await Boot.create(
        Module(
            "/",
            Controllers=[WsCtrl]
        ),
        api_version=ApiVersion(supported={1, 2, 3}),
        global_websocket_route="/ws/v{version}"
    )

    wsctrl: WsCtrl = validation.apply(Di.ie().find("WsCtrl"), WsCtrl)

    assert wsctrl.final_routes == {
        "/ws/v2", "/ws/v3", "/ws/v2/connect", "/ws/v3/connect"
    }



import pytest
from fastapi import Header

from orwynn.base.module.module import Module
from orwynn.boot import Boot
from orwynn.http import Endpoint, HttpController
from orwynn.testing.client import Client
from orwynn.testingtools import HeadersGetController


class _Ctrl1(HttpController):
    Route = "/"
    Endpoints = [
        Endpoint(method="get")
    ]

    def get(self, x_testing: str | None = Header(default=None)) -> dict:
        return {
            "x-testing": x_testing or None
        }


class _Ctrl2(HttpController):
    Route = "/"
    Endpoints = [
        Endpoint(method="get")
    ]

    def get(
        self,
        x_testing: str | None = Header(default=None),
        x_tmp: str | None = Header(default=None)
    ) -> dict:
        return {
            "x-testing": x_testing or None,
            "x-tmp": x_tmp or None
        }


@pytest.mark.asyncio
async def test_bind_headers():
    boot: Boot = await Boot.create(
        Module("/", Controllers=[_Ctrl1])
    )

    binded: Client = boot.app.client.bind_headers({
        "x-testing": "hello"
    })

    data: dict = binded.get_jsonify("/")
    assert data["x-testing"] == "hello"


@pytest.mark.asyncio
async def test_bind_headers_accumulate():
    boot: Boot = await Boot.create(
        Module("/", Controllers=[_Ctrl2])
    )

    binded: Client = boot.app.client.bind_headers({
        "x-testing": "hello"
    }).bind_headers({
        "x-tmp": "world"
    })

    data: dict = binded.get_jsonify("/")
    assert data["x-testing"] == "hello"
    assert data["x-tmp"] == "world"

@pytest.mark.asyncio
async def test_multiple_bind_headers():
    """
    Shouldn't stack bind headers in the original client for subsequent
    bind_headers() calls.
    """
    boot: Boot = await Boot.create(
        Module("/", Controllers=[HeadersGetController])
    )

    binded_1: Client = boot.app.client.bind_headers({
        "x-testing": "hello"
    })

    binded_2: Client = binded_1.bind_headers({
        "x-tmp": "world"
    })

    assert binded_1.binded_headers == {"x-testing": "hello"}
    assert binded_2.binded_headers == {
        "x-testing": "hello",
        "x-tmp": "world"
    }



from orwynn.http.context.id import HttpRequestContextId
from orwynn.http.middleware.builtinmiddleware import BuiltinHttpMiddleware
from orwynn.http.middleware.nextcall import HttpNextCall
from orwynn.http.requests import HttpRequest
from orwynn.http.responses import HttpResponse
from orwynn.log import Log


class HttpRequestContextBuiltinMiddleware(BuiltinHttpMiddleware):
    """Populates the context storage with the current request's id."""
    async def process(
        self, request: HttpRequest, call_next: HttpNextCall
    ) -> HttpResponse:
        request_id: str = HttpRequestContextId().save()

        # Also contextualize logs
        with Log.contextualize(**{"http.request_id": request_id}):
            return await call_next(request)



from collections.abc import Callable
from typing import Any

from orwynn.base import Model


class FuncSpec(Model):
    """
    Holds information about some function.
    """
    fn: Callable
    args: tuple | None = None
    kwargs: dict | None = None

    def call(
        self,
        *,
        prepended_extra_args: tuple | None = None,
        appended_extra_args: tuple | None = None,
        prepended_extra_kwargs: dict | None = None,
        appended_extra_kwargs: dict | None = None,
    ) -> Any:
        """
        Calls a specified in this spec function.
        """
        args: tuple = ()
        if self.args is not None:
            args = self.args

        kwargs: dict = {}
        if self.kwargs is not None:
            kwargs = self.kwargs

        if prepended_extra_args is None:
            prepended_extra_args = ()
        if appended_extra_args is None:
            appended_extra_args = ()
        if prepended_extra_kwargs is None:
            prepended_extra_kwargs = {}
        if appended_extra_kwargs is None:
            appended_extra_kwargs = {}

        final_args: tuple = prepended_extra_args + args + appended_extra_args
        final_kwargs: dict = {
            **prepended_extra_kwargs,
            **kwargs,
            **appended_extra_kwargs,
        }

        return self.fn(*final_args, **final_kwargs)



import pytest

from orwynn.base.module.module import Module
from orwynn.boot.boot import Boot
from orwynn.http.controller.controller import HttpController
from orwynn.http.controller.endpoint.endpoint import Endpoint
from orwynn.testing.client import Client


@pytest.mark.asyncio
async def test_basic():
    class C1(HttpController):
        Route = "/"
        Endpoints = [Endpoint(method="get")]

        def get(self):
            return {}

    boot: Boot = await Boot.create(
        Module(route="/", Controllers=[C1]),
        apprc={
            "prod": {
                "App": {
                    "cors": {
                        "is_enabled": True,
                        "allow_origins": ["*"]
                    }
                }
            }
        }
    )
    http: Client = boot.app.client

    r = http.options(
        "/",
        headers={
            "origin": "hello",
            "access-control-request-method": "GET"
        }
    )

    assert r.headers.get("access-control-allow-origin") == "*"


@pytest.mark.asyncio
async def test_correct_origin():
    class C1(HttpController):
        Route = "/"
        Endpoints = [Endpoint(method="get")]

        def get(self) -> dict:
            return {}

    boot: Boot = await Boot.create(
        Module(route="/", Controllers=[C1]),
        apprc={
            "prod": {
                "App": {
                    "cors": {
                        "is_enabled": True,
                        "allow_origins": ["hello"]
                    }
                }
            }
        }
    )
    http: Client = boot.app.client


    r = http.options(
        "/",
        headers={
            "origin": "hello",
            "access-control-request-method": "POST"
        }
    )

    assert r.headers.get("access-control-allow-origin") == "hello"


@pytest.mark.asyncio
async def test_wrong_origin():
    class C1(HttpController):
        Route = "/"
        Endpoints = [Endpoint(method="get")]

    boot: Boot = await Boot.create(
        Module(route="/", Controllers=[C1]),
        apprc={
            "prod": {
                "App": {
                    "cors": {
                        "is_enabled": True,
                        "allow_origins": ["nothello"]
                    }
                }
            }
        }
    )
    http: Client = boot.app.client

    r = http.options(
        "/",
        headers={
            "origin": "hello",
            "access-control-request-method": "POST"
        }
    )

    assert r.headers.get("access-control-allow-origin") is None


@pytest.mark.asyncio
async def test_unsuccessful():
    """
    Unsuccessful responses should also contain according CORS headers.
    """
    class C1(HttpController):
        Route = "/"
        Endpoints = [Endpoint(method="get")]

        def get(self):
            raise ValueError("cors test")

    boot: Boot = await Boot.create(
        Module(route="/", Controllers=[C1]),
        apprc={
            "prod": {
                "App": {
                    "cors": {
                        "is_enabled": True,
                        "allow_origins": ["*"]
                    }
                }
            }
        }
    )
    http: Client = boot.app.client

    r = http.options(
        "/",
        headers={
            "origin": "hello",
            "access-control-request-method": "GET"
        }
    )
    assert r.headers.get("access-control-allow-origin") == "*"

    r = http.get(
        "/",
        headers={
            "origin": "hello"
        }
    )
    assert r.json()["type"] == "error"
    assert r.headers.get("access-control-allow-origin") == "*"



import functools
from typing import Any, Callable

from orwynn.utils.types import TClass

from .errors import ClassNotFoundError


class Static:
    def __init__(self) -> None:
        raise NotImplementedError


class ClassUtils(Static):
    @classmethod
    def find_all_subclasses(
        cls,
        Base: type[TClass],
    ) -> list[type[TClass]]:
        """
        Recursively searches for all subclasses of the base class and returns
        it.
        """
        Classes: list[type[TClass]] = []

        cls._traverse_for_subclasses(Base, Classes)

        return Classes

    @classmethod
    def _traverse_for_subclasses(
        cls,
        Start: type[TClass],
        target_list: list[type[TClass]],
    ):
        for klass in Start.__subclasses__():  # type: ignore
            target_list.append(klass)
            cls._traverse_for_subclasses(klass, target_list)

    @classmethod
    def find_subclass_by_name(cls, name: str, BaseClass: TClass) -> TClass:
        """Searches given base class for subclass with given name.

        Args:
            name:
                Name of the target class.
            BaseClass:
                Where to search for subclasses.

        Returns:
            Class found.
        """
        # Maybe given class has correct name
        if BaseClass.__name__ == name:
            return BaseClass

        out: TClass | None = cls._traverse_subclasses_checking_name(
            name, BaseClass
        )

        if out is None:
            raise ClassNotFoundError(
                f"class of supertype {BaseClass} with name {name} is not found"
            )
        else:
            return out

    @classmethod
    def bind_first_arg(cls, arg: Any):
        """Adds first argument to wrapped function call.

        Useful when you want to add e.g. "self" with some instance to
        dynamically obtained class's method.

        Args:
            arg:
                First arg to be added to wrapped function call.

        Returns:
            Wrapper accepting function.
        """
        def wrapper(fn: Callable):
            @functools.wraps(fn)
            def inner(*args, **kwargs):
                return fn(arg, *args, **kwargs)
            return inner
        return wrapper

    @classmethod
    def bind_first_arg_async(cls, arg: Any):
        """Same as bind_first_arg(), but for async functions.
        """
        def wrapper(fn: Callable):
            @functools.wraps(fn)
            async def inner(*args, **kwargs):
                return await fn(arg, *args, **kwargs)
            return inner
        return wrapper

    @classmethod
    def _traverse_subclasses_checking_name(
        cls,
        name: str,
        C: TClass
    ) -> TClass | None:
        for SubClass in C.__subclasses__():
            if SubClass.__name__ == name:
                return SubClass
            else:
                result: TClass | None = \
                    cls._traverse_subclasses_checking_name(name, SubClass)
                if result is not None:
                    return result

        return None



from orwynn.singleton.singleton import Singleton


class Worker(Singleton):
    """Does framework-related tasks, such as assembling of all app or DI."""
    def __init__(self) -> None:
        super().__init__()



"""Cryptographic operations.

For password hashing uses bcrypt, which has the salt saved into the hash
itself.

See password hashing https://stackoverflow.com/a/23768422/14748231
"""
import bcrypt

from orwynn.utils import validation


def hash_password(plain_password: str, encoding: str = "utf-8") -> str:
    validation.validate(plain_password, str)
    validation.validate(encoding, str)

    return bcrypt.hashpw(
        plain_password.encode(encoding),
        bcrypt.gensalt()
    ).decode(encoding)


def check_password(
    plain_password: str,
    hashed_password: str,
    encoding: str = "utf-8"
) -> bool:
    validation.validate(plain_password, str)
    validation.validate(hashed_password, str)
    validation.validate(encoding, str)

    return bcrypt.checkpw(
        plain_password.encode(encoding),
        hashed_password.encode(encoding)
    )




import pytest
from fastapi import Query, Request

from orwynn.apiversion import ApiVersion
from orwynn.base.controller.errors import (
    AlreadyRegisteredMethodError,
    MissingControllerClassAttributeError,
)
from orwynn.base.model import Model
from orwynn.base.module import Module
from orwynn.boot import Boot
from orwynn.di.di import Di
from orwynn.helpers.web import REQUEST_METHOD_BY_PROTOCOL
from orwynn.http import Endpoint, HttpController
from orwynn.http.controller.errors import DefinedTwiceControllerMethodError
from orwynn.http.errors import HttpException, UnsupportedHttpMethodError
from orwynn.proxy.boot import BootProxy
from orwynn.testing import Client
from orwynn.utils import validation
from orwynn.utils.scheme import Scheme
from orwynn.utils.validation import validate_re
from orwynn.utils.validation.errors import (
    RequestValidationException,
    ReValidationError,
    ValidationError,
)
from tests.std.text import DEFAULT_ID, Text


def test_http_methods():
    for method in REQUEST_METHOD_BY_PROTOCOL[Scheme.HTTP]:
        assert hasattr(HttpController, method.value)


@pytest.mark.asyncio
async def test_undefined_route():
    class C1(HttpController):
        Endpoints = [Endpoint(method="get")]

    m1 = Module(route="/", Controllers=[C1])
    await validation.expect_async(
        Boot.create(m1),
        MissingControllerClassAttributeError
    )


@pytest.mark.asyncio
async def test_invalid_route():
    class C1(HttpController):
        Route = "i don't like rules"
        Endpoints = [Endpoint(method="get")]

    m1 = Module(route="/", Controllers=[C1])
    await validation.expect_async(
        Boot.create(m1),
        ReValidationError
    )


@pytest.mark.asyncio
async def test_undefined_endpoints():
    class C1(HttpController):
        Route = "/c1"

    m1 = Module(route="/", Controllers=[C1])
    await validation.expect_async(
        Boot.create(m1),
        MissingControllerClassAttributeError
    )


@pytest.mark.asyncio
async def test_empty_endpoints():
    class C1(HttpController):
        Route = "/c1"
        Endpoints = []

    m1 = Module(route="/", Controllers=[C1])
    await validation.expect_async(
        Boot.create(m1),
        ValidationError
    )


@pytest.mark.asyncio
async def test_unsupported_method():
    class C1(HttpController):
        Route = "/c1"
        Endpoints = [Endpoint(method="donuts")]

    m1 = Module(route="/", Controllers=[C1])
    await validation.expect_async(
        Boot.create(m1),
        UnsupportedHttpMethodError
    )


@pytest.mark.asyncio
async def test_defined_twice_method():
    class C1(HttpController):
        Route = "/c1"
        Endpoints = [Endpoint(method="get"), Endpoint(method="get")]

    m1 = Module(route="/", Controllers=[C1])
    await validation.expect_async(
        Boot.create(m1),
        DefinedTwiceControllerMethodError
    )


@pytest.mark.asyncio
async def test_uppercase_methods():
    class C1(HttpController):
        Route = "/c1"
        Endpoints = [
            Endpoint(method="GET"),
            Endpoint(method="POST")
        ]

    m1 = Module(route="/", Controllers=[C1])
    await Boot.create(m1)


@pytest.mark.asyncio
async def test_already_registered():
    class C1(HttpController):
        Route = "/hello"
        Endpoints = [Endpoint(method="get")]

    class C2(HttpController):
        Route = "/hello"
        Endpoints = [Endpoint(method="get")]

    m1 = Module(route="/", Controllers=[C1, C2])
    await validation.expect_async(
        Boot.create(m1),
        AlreadyRegisteredMethodError
    )


def test_std_routes(std_boot: Boot, std_http: Client):
    json: dict = std_http.get_jsonify("/text")
    text: Text = Text.recover(json)
    validate_re(text.text, DEFAULT_ID + r"\: .+")


@pytest.mark.asyncio
async def test_default_404():
    class C1(HttpController):
        Route = "/"
        Endpoints = [Endpoint(method="get")]

    boot: Boot = await Boot.create(
        Module(route="/", Controllers=[C1])
    )

    data: dict = boot.app.client.get_jsonify(
        "/pizza",
        404
    )

    recovered_exception: HttpException = validation.apply(
        BootProxy.ie().api_indication.recover(
            HttpException,
            data
        ),
        HttpException
    )

    assert recovered_exception.status_code == 404
    assert recovered_exception.detail == "Not Found"


@pytest.mark.asyncio
async def test_default_request_validation_error():
    class Item(Model):
        name: str
        price: float

    class C1(HttpController):
        Route = "/"
        Endpoints = [Endpoint(method="post")]

        def post(self, item: Item) -> dict:
            return {}

    data: dict = (await Boot.create(
        Module(route="/", Controllers=[C1])
    )).app.client.post_jsonify(
        "/",
        422,
        json={
            "name": 222,
            "price": "hello"
        }
    )

    # Temporarily content of this exception is not checked since it is not
    # filled back to the indication
    validation.apply(
        BootProxy.ie().api_indication.recover(
            RequestValidationException,
            data
        ),
        RequestValidationException
    )


@pytest.mark.asyncio
async def test_default_method_not_allowed():
    class C1(HttpController):
        Route = "/"
        Endpoints = [Endpoint(method="get")]

        def get(self) -> dict:
            return {}

    data: dict = (await Boot.create(
        Module(route="/", Controllers=[C1])
    )).app.client.post_jsonify(
        "/",
        405
    )

    recovered_exception: HttpException = validation.apply(
        BootProxy.ie().api_indication.recover(
            HttpException,
            data
        ),
        HttpException
    )

    assert recovered_exception.status_code == 405
    assert recovered_exception.detail == "Method Not Allowed"


@pytest.mark.asyncio
async def test_final_routes():
    class _Ctrl(HttpController):
        Route = "/{id}/tasty"
        Endpoints = [
            Endpoint(method="get")
        ]

    await Boot.create(
        Module("/donuts", Controllers=[_Ctrl]),
        global_http_route="/api/v{version}"
    )

    ctrl: HttpController = validation.apply(
        Di.ie().find("_Ctrl"),
        HttpController
    )

    assert "/api/v1/donuts/{id}/tasty" in ctrl.final_routes


@pytest.mark.asyncio
async def test_is_matching_route():
    class _Ctrl(HttpController):
        Route = "/{id}/tasty"
        Endpoints = [
            Endpoint(method="get")
        ]
        Version = 2

    await Boot.create(
        Module("/donuts", Controllers=[_Ctrl]),
        global_http_route="/api/v{version}",
        api_version=ApiVersion(supported={1, 2, 3})
    )

    ctrl: HttpController = validation.apply(
        Di.ie().find("_Ctrl"),
        HttpController
    )

    assert ctrl.is_matching_route("/api/v2/donuts/e67840v/tasty") is True

    # Another versions should not be listed if a controller does not support
    # them
    assert ctrl.is_matching_route("/api/v1/donuts/e67840v/tasty") is False
    assert ctrl.is_matching_route("/api/v3/donuts/helloworld/tasty") is False
    ##

    assert ctrl.is_matching_route("/api/v2/donuts/e67840v/tasty/gogo") is False
    assert ctrl.is_matching_route("/api/v2/donuts/tasty") is False
    assert ctrl.is_matching_route("/api/donuts/eb00v/tasty") is False
    assert ctrl.is_matching_route("/donuts/eb00v/tasty") is False


@pytest.mark.asyncio
async def test_multiple_query_params():
    """
    Should correctly parse list of query params.
    """
    class _Ctrl(HttpController):
        Route = "/items"
        Endpoints = [
            Endpoint(method="get")
        ]

        def get(
            self,
            request: Request,
            q: list[str] | None = Query(None)
        ) -> dict:
            return {
                "q": q
            }

    boot: Boot = await Boot.create(
        Module("/", Controllers=[_Ctrl])
    )

    data: dict = boot.client.get_jsonify(
        "/items?q=1&q=2",
        200
    )

    assert data == {
        "q": ["1", "2"]
    }



from orwynn.mapping.errors import UnsetIdMappingError
from orwynn.utils import validation

from .mapping import Mapping


class _M1(Mapping):
    number: int


def test_try_set_id():
    m: Mapping = _M1(id="helloworld", number=1)

    assert m.id == "helloworld"


def test_get_id():
    m: Mapping = _M1(id="helloworld", number=2)

    assert m.getid() == "helloworld"


def test_get_id_unset():
    m: Mapping = _M1(number=2)

    validation.expect(
        m.getid,
        UnsetIdMappingError
    )



import typing
from enum import Enum
from typing import TYPE_CHECKING, Any

from orwynn.mongo.errors import (
    MongoTypeConversionError,
    UnsetIdMongoError,
)
from orwynn.mongo.types import MongoCompatibleType, TDocument
from orwynn.utils.klass import Static

if TYPE_CHECKING:
    from orwynn.mongo import Document
    from orwynn.mongo.search import DocumentSearch

MongoCompatibleTypes: tuple[Any, ...] = typing.get_args(MongoCompatibleType)


class MongoUtils(Static):
    @staticmethod
    def process_query(
        query: dict[str, Any],
        search: "DocumentSearch[TDocument]",
        DocumentClass: type[TDocument],
        *,
        find_all_kwargs: dict | None = None,
    ) -> list[TDocument]:
        if find_all_kwargs is None:
            find_all_kwargs = {}
        result: list[TDocument] = list(DocumentClass.find_all(
            query,
            **find_all_kwargs,
        ))

        if len(result) == 0:
            raise search.get_not_found_error("TDocument")
        if search.expectation is not None:
            search.expectation.check(result)

        return result


    @staticmethod
    def query_by_nested_dict(
        query: dict[str, Any],
        nested_dict: dict[str, Any],
        root_key: str,
    ) -> None:
        """
        Updates query for searching nested dict values.

        Args:
            query:
                Query to update.
            nested_dict:
                Data to search.
            root_key:
                Outermost key of field containing the nested dict.

        Example:
        ```python
        class MyDocument(Document):
            mydata: dict

        query = {}
        nd = {
            "mycode": {
                "approximate": {
                    "$in": [12, 34]
                }
            }
        }
        root_key = "mydata"

        MongoUtils.query_by_nested_dict(
            query,
            nd,
            root_key
        )
        # query = {"mydata.mycode.approximate": {"$in": [12, 34]}}
        ```
        """
        converted_data: dict[str, Any] = MongoUtils.convert_dict({
            root_key: nested_dict,
        })
        query.update(converted_data)

    @staticmethod
    def convert_dict(d: dict[str, Any]) -> dict[str, Any]:
        """
        Converts dictionary into a Mongo search format.

        All key structures is converted to dot-separated string, e.g.
        `{"key1": {"key2": {"key3_1": 10, "key3_2": 20}}}` is converted to
        `{"key1.key2.key3_1": 10, "key1.key2.key3_2": 20}`.

        Keys started with dollar sign are not converted and left as it is:
        ```python
        # input
        {
            "a1": {
                "a2": {
                    "$in": my_list
                }
            }
        }

        # output
        {
            "a1.a2": {
                "$in": my_list
            }
        }
        ```
        """
        result: dict[str, Any] = {}

        for k, v in d.items():

            if k.startswith("$") or not isinstance(v, dict):
                result[k] = v
                continue

            for k1, v1 in MongoUtils.convert_dict(v).items():
                if k1.startswith("$"):
                    result[k] = {
                        k1: v1,
                    }
                    continue
                result[k + "." + k1] = v1

        return result


    @staticmethod
    def get_id(document: "Document") -> str:
        doc_id: str | None = document.id

        if doc_id is None:
            raise UnsetIdMongoError

        return doc_id

    @staticmethod
    def convert_compatible(obj: Any) -> MongoCompatibleType:
        """
        Converts object to mongo compatible type.

        Convertation rules:
        - object with type listed in already compatible mongo types is returned
        as it is
        - elements of list, as well as dictionary's keys and values are
        converted recursively using this function
        - in case of Enum, the Enum's value is obtained and converted through
        this function
        - objects with defined attribute `mongovalue` (either by variable or
        property) is called like `obj.mongovalue` and the result is converted
        again through this function
        - for all other types the MongoTypeConversionError is raised

        Args:
            obj:
                Object to convert.

        Raises:
            MongoTypeConversionError:
                Cannot convert object to mongo-compatible.
        """
        result: MongoCompatibleType

        if type(obj) is dict:
            result = {}
            for k, v in obj.items():
                result[MongoUtils.convert_compatible(k)] = \
                    MongoUtils.convert_compatible(v)
        elif type(obj) is list:
            result = []
            for item in obj:
                result.append(MongoUtils.convert_compatible(item))
        elif type(obj) in MongoCompatibleTypes:
            result = obj
        elif hasattr(obj, "mongovalue"):
            result = MongoUtils.convert_compatible(obj.mongovalue)
        elif isinstance(obj, Enum):
            result = MongoUtils.convert_compatible(obj.value)
        else:
            raise MongoTypeConversionError(t=type(obj))

        return result

    @staticmethod
    def refresh(document: TDocument) -> TDocument:
        """
        Refreshes document with fresh data from database.
        """
        return document.find_one({"id": document.getid()})



from orwynn.websocket.middleware.builtin import (
    BuiltinWebsocketMiddleware,
)
from orwynn.websocket.middleware.nextcall import WebsocketNextCall
from orwynn.websocket.websocket import Websocket


class ConnectionBuiltinWebsocketMiddleware(BuiltinWebsocketMiddleware):
    """Manages websocket's connection acceptance.

    Note that you don't need to use websocket.accept() anywhere in your program
    or you will stuck in an endless loop. But you can call websocket.close()
    whenever you want (multiple calls don't raise any exceptions).

    Maybe, in future there will be a possibility to disable this Middleware
    (and thus accept a connection everywhere you want).
    """

    async def process(
        self, request: Websocket, call_next: WebsocketNextCall
    ) -> None:
        await request.accept()
        await call_next(request)



from orwynn.base.model import Model
from orwynn.http import (
    Endpoint,
    HttpController,
    HttpRequest,
    RedirectHttpResponse,
)

GET_DATA: dict = {"message": "hello"}


class GetController(HttpController):
    Route = "/"
    Endpoints = [
        Endpoint(method="get")
    ]

    def get(self) -> dict:
        return GET_DATA


class IdGetController(HttpController):
    Route = "/{id}"
    Endpoints = [
        Endpoint(method="get")
    ]

    def get(self, id: str) -> dict:
        return {"id": id}


class HeadersGetController(HttpController):
    """
    Returns all accepted headers.
    """
    Route = "/"
    Endpoints = [
        Endpoint(method="get")
    ]

    def get(self, request: HttpRequest) -> dict:
        return {k: v for k, v in request.headers.items()}


class RedirectController(HttpController):
    Route = "/"
    Endpoints = [
        Endpoint(
            method="get"
        )
    ]

    def get(self) -> RedirectHttpResponse:
        return RedirectHttpResponse("https://google.com")


class Item(Model):
    name: str
    price: float
    order: int = 0



import pytest

from orwynn.base.module.module import Module
from orwynn.boot.boot import Boot
from orwynn.testing.client import Client
from orwynn.websocket import (
    Websocket,
    WebsocketController,
    WebsocketMiddleware,
    WebsocketNextCall,
)


class Mw1(WebsocketMiddleware):
    async def process(
        self, request: Websocket, call_next: WebsocketNextCall
    ) -> None:
        await request.send_json({"value": "entry"})
        await call_next(request)


class Ws1(WebsocketController):
    Route = "/"

    async def on_message(self, websocket: Websocket) -> None:
        await websocket.send_json({"value": "hello"})
        await websocket.close()


@pytest.mark.asyncio
async def test_basic():
    boot: Boot = await Boot.create(Module(
        route="/hello",
        Controllers=[Ws1],
        Middleware=[Mw1]
    ))
    client: Client = boot.app.client

    with client.websocket("/hello/message") as ws:
        data: dict

        data = ws.receive_json()
        assert data == {"value": "entry"}

        data = ws.receive_json()
        assert data == {"value": "hello"}



import pytest

from orwynn.base.error.errors import ExceptionAlreadyHandledError
from orwynn.base.errorhandler import ErrorHandler
from orwynn.base.module import Module
from orwynn.base.service.service import Service
from orwynn.boot.boot import Boot
from orwynn.http.context.middleware.contextbuiltin import (
    HttpRequestContextBuiltinMiddleware,
)
from orwynn.http.controller.controller import HttpController
from orwynn.http.controller.endpoint.endpoint import Endpoint
from orwynn.http.middleware.middleware import HttpMiddleware
from orwynn.http.middleware.nextcall import HttpNextCall
from orwynn.http.requests import HttpRequest
from orwynn.http.responses import (
    HttpResponse,
    JsonHttpResponse,
    TestHttpResponse,
)
from orwynn.proxy.boot import BootProxy
from orwynn.testing import Client
from orwynn.testingtools import Item
from orwynn.utils import validation


class GeneralEh(ErrorHandler):
    E = Exception

    def handle(
        self, request: HttpRequestContextBuiltinMiddleware, error: Exception
    ):
        return JsonHttpResponse(
            BootProxy.ie().api_indication.digest(error),
            401
        )


class RaiseErrorController(HttpController):
    Route = "/"
    Endpoints = [Endpoint(method="get")]

    def get(self):
        raise ValueError("hello")


@pytest.mark.asyncio
async def test_custom_handler():
    class C1(HttpController):
        Route = "/"
        Endpoints = [Endpoint(method="get")]

        def get(self):
            raise ValueError("whoops!")

    boot: Boot = await Boot.create(
        Module(route="/", Controllers=[C1]),
        ErrorHandlers={GeneralEh}
    )
    http: Client = boot.app.client

    r: TestHttpResponse = http.get("/", 401)

    recovered_error: ValueError = validation.apply(
        BootProxy.ie().api_indication.recover(ValueError, r.json()),
        ValueError
    )

    assert recovered_error.args[0] == "whoops!"


@pytest.mark.asyncio
async def test_default_exception():
    class C1(HttpController):
        Route = "/"
        Endpoints = [Endpoint(method="get")]

        def get(self):
            raise TypeError("whoops!")

    boot: Boot = await Boot.create(
        Module(route="/", Controllers=[C1])
    )
    http: Client = boot.app.client

    r: TestHttpResponse = http.get("/", 400)

    recovered_exception: TypeError = validation.apply(
        BootProxy.ie().api_indication.recover(TypeError, r.json()),
        TypeError
    )

    assert recovered_exception.args[0] == "whoops!"


@pytest.mark.asyncio
async def test_as_acceptor():
    """
    Tests if an error handler can truly accept Providers.
    """
    ASSERTED_TEXT: str = "Hello, world!"

    class CoolService(Service):
        def do_something(
            self
        ) -> str:
            return ASSERTED_TEXT

    class C1(HttpController):
        Route = "/"
        Endpoints = [Endpoint(method="get")]

        def get(self):
            raise TypeError("whoops!")

    class Eh1(ErrorHandler):
        E = Exception

        def __init__(self, cool_service: CoolService) -> None:
            self.__cool_service: CoolService = cool_service

        def handle(
            self,
            request: HttpRequest,
            error: Exception
        ) -> HttpResponse:
            data: dict = BootProxy.ie().api_indication.digest(error)
            data["__test_meta_info"] = self.__cool_service.do_something()
            return JsonHttpResponse(data, 400)

    boot: Boot = await Boot.create(
        Module(route="/", Providers=[CoolService], Controllers=[C1]),
        ErrorHandlers={Eh1}
    )
    client: Client = boot.app.client

    data: dict = client.get_jsonify("/", 400)

    assert data["__test_meta_info"] == ASSERTED_TEXT


@pytest.mark.asyncio
async def test_default_in_middleware():
    """
    Should handle exceptions occured in middleware.
    """
    class C1(HttpController):
        Route = "/"
        Endpoints = [Endpoint(method="get")]

        def get(self):
            return {}

    class M1(HttpMiddleware):
        async def process(
            self,
            request: HttpRequest,
            call_next: HttpNextCall
        ) -> HttpResponse:
            raise ValueError("whoops!")

    boot: Boot = await Boot.create(
        Module(route="/", Controllers=[C1], Middleware=[M1])
    )

    r: TestHttpResponse = boot.app.client.get("/", 400)

    recovered_exception: Exception = validation.apply(
        BootProxy.ie().api_indication.recover(Exception, r.json()),
        Exception
    )

    assert recovered_exception.args[0] == "whoops!"


@pytest.mark.asyncio
async def test_custom_in_middleware():
    """
    Should handle exceptions occured in middleware.
    """
    class C1(HttpController):
        Route = "/"
        Endpoints = [Endpoint(method="get")]

        def get(self):
            return {}

    class M1(HttpMiddleware):
        async def process(
            self,
            request: HttpRequest,
            call_next: HttpNextCall
        ) -> HttpResponse:
            raise ValueError("whoops!")

    boot: Boot = await Boot.create(
        Module(route="/", Controllers=[C1], Middleware=[M1]),
        ErrorHandlers={GeneralEh}
    )

    r: TestHttpResponse = boot.app.client.get("/", 401)

    recovered: ValueError = validation.apply(
        BootProxy.ie().api_indication.recover(ValueError, r.json()),
        ValueError
    )

    assert recovered.args[0] == "whoops!"


@pytest.mark.asyncio
async def test_exception_handled_twice():
    """
    Should raise an error for twice-handled exceptions.
    """
    class Eh1(ErrorHandler):
        E = Exception

    await validation.expect_async(
        Boot.create(
            Module("/", Controllers=[RaiseErrorController]),
            ErrorHandlers={GeneralEh, Eh1}
        ),
        ExceptionAlreadyHandledError
    )


@pytest.mark.asyncio
async def test_pydantic_validation_error_is_catched():
    """
    Should catch pydantic.ValidationError too, despite it's internal multiple
    inheritance.
    """
    class _Ctrl(HttpController):
        Route = "/"
        Endpoints = [
            Endpoint(method="get")
        ]

        def get(self) -> dict:
            # Invoke a validation error
            Item(name="item", price="whocares")  # type: ignore
            return {}

    boot: Boot = await Boot.create(
        Module("/", Controllers=[_Ctrl])
    )

    boot.app.client.get_jsonify("/", 400)






class DatabaseEntityNotFoundError(Exception):
    """When requested database entity is not found."""
    def __init__(
        self,
        message: str = "",
        *,
        collection: str | None = None,
        query: dict | None = None
    ) -> None:
        if not message and collection and query:
            message = \
                f"not found entity for query {query} in collection" \
                + f" {collection}"
        super().__init__(message)


class UnknownDatabaseKindError(Exception):
    pass



from orwynn.base.module.module import Module
from orwynn.sql.search import TableSearch
from orwynn.sql.shd import SHD
from orwynn.sql.stateflag import StateFlag
from orwynn.sql.types import ConvertedModel, ListedConvertedModel, TTable
from orwynn.sql.utils import SHDUtils, SQLUtils

from .config import SQLConfig
from .sql import SQL
from .table import Table

__all__ = [
    "Table",
    "SQL",
    "SQLConfig",
    "SHDUtils",
    "SQLUtils",
    "StateFlag",
    "TTable",
    "ConvertedModel",
    "ListedConvertedModel",
    "SHD",
    "TableSearch"
]

module = Module(
    Providers=[SQL, SQLConfig],
    exports=[SQL, SQLConfig]
)



from orwynn.utils import validation
from orwynn.utils.condition import ComparisonCondition, ComparisonMark
from orwynn.utils.expectation import ListExpectation
from orwynn.utils.expectation.errors import (
    ExpectationError,
    UnsupportedExpectationTypeError,
)


def test_count():
    expectation: ListExpectation = ListExpectation(count=ComparisonCondition(
        mark=ComparisonMark.MoreEqual,
        value=3,
    ))

    validation.expect(
        expectation.check,
        ExpectationError,
        [1, 2],
    )
    expectation.check([1, 2, 3])
    expectation.check([1, 2, 3, 4])


def test_unsupported_type():
    expectation: ListExpectation = ListExpectation(count=ComparisonCondition(
        mark=ComparisonMark.MoreEqual,
        value=2,
    ))

    validation.expect(
        expectation.check,
        UnsupportedExpectationTypeError,
        # only list should be expected as a target
        2,
    )



from typing import Generic

from orwynn.mongo.document import Document
from orwynn.mongo.types import TDocument
from orwynn.utils.dbsearch import DatabaseSearch


class DocumentSearch(DatabaseSearch[Document], Generic[TDocument]):
    """
    Search Mongo Documents.
    """


class MongoStateFlagSearch(DocumentSearch):
    keys: list[str] | None = None
    values: list[bool] | None = None



from typing import Any, ClassVar, Generic

from antievil import AbstractUsageError
from pydantic.generics import GenericModel

from orwynn.utils.condition import ComparisonCondition as _ComparisonCondition
from orwynn.utils.expectation.errors import (
    ExpectationError,
    UnsupportedExpectationTypeError,
)
from orwynn.utils.types import T


class Expectation(GenericModel, Generic[T]):
    """
    Parameter-model signifies which checks should be performed.

    All expectations are based on return type. E.g. if your function returns
    a List, you should consider using ListExpectation.

    All expectations can be checked via general function `self.check()`.

    Class-attributes:
        SUPPPORTED_TYPES(optional):
            Which types the expectation supports. Defaults to None, but only
            for abstract classes.

    @abstract
    """
    SUPPORTED_TYPES: ClassVar[list[type] | None] = None

    @classmethod
    def is_supported(cls, Type: type[T]) -> bool:
        """
        Checks whether the given type is supported by the class.
        """
        if cls.SUPPORTED_TYPES is None:
            raise AbstractUsageError(
                explanation="cannot check supported types",
                Class=cls,
            )
        else:
            return Type in cls.SUPPORTED_TYPES

    def check(
        self,
        target: T,
    ) -> None:
        """
        Checks expectation against the given target.

        Args:
            target:
                Target to check.

        Raises:
            UnsupportedExpectationTypeError:
                The type of given target is not supported by the expectation.
            ExpectationError:
                Any of expectation checks against the target has been failed.
        """
        self._check_target_type(target)
        self._check(target)

    def _check(self, target: T):
        """
        Main check function. Should be redefined at children.
        """
        raise AbstractUsageError(
            explanation="cannot call check function",
            Class=type(self),
        )

    def _check_target_type(self, target: T) -> None:
        if not self.is_supported(type(target)):
            raise UnsupportedExpectationTypeError(
                UnsupportedType=type(target),
                expectation=self,
            )

    class Config:
        arbitrary_types_allowed = True


class ListExpectation(Expectation[T], Generic[T]):
    """
    Expectation rules for List return type.

    Attributes:
        count(optional):
            Condition signifies how many items should the list contain.
    """
    SUPPORTED_TYPES = [list]

    count: _ComparisonCondition[int] | None = None

    def _check(self, target: list[T]):
        if self.count is not None and not self.count.compare(len(target)):
            raise ExpectationError(
                title="count",
                actual_value=len(target),
                expected_value=self.count,
            )



from orwynn.base.errorhandler.errorhandler import ErrorHandler
from orwynn.proxy.boot import BootProxy
from orwynn.websocket.websocket import Websocket


class DefaultWebsocketErrorHandler(ErrorHandler):
    E = Exception

    async def handle(
        self, request: Websocket, error: Exception
    ) -> None:
        await request.send_json(BootProxy.ie().api_indication.digest(
            error
        ))



from typing import Generic

from antievil import (
    NotFoundError,
)
from pydantic.generics import GenericModel

from orwynn.utils.expectation import Expectation
from orwynn.utils.types import T


class DatabaseSearch(GenericModel, Generic[T]):
    """
    Search terms to find database's objects.

    Should be subclassed in order to add extra search terms and specific
    database type-related fields.

    @abstract
    """
    ids: list[str] | None = None
    expectation: Expectation | None = None

    def get_not_found_error(self, title: str) -> NotFoundError:
        return NotFoundError(
            title=title,
            options=self.dict(),
        )



from orwynn.base.middleware import Middleware
from orwynn.websocket.middleware.nextcall import WebsocketNextCall
from orwynn.websocket.websocket import Websocket


class WebsocketMiddleware(Middleware):
    """Intermediate operational layer for Websocket requests.

    Note that websocket methods return None since websocket sends data through
    the Websocket object itself.
    """
    async def dispatch(
        self,
        request: Websocket,
        call_next: WebsocketNextCall
    ) -> None:
        return await super().dispatch(request, call_next)

    async def process(
        self,
        request: Websocket,
        call_next: WebsocketNextCall
    ) -> None:
        return await super().process(request, call_next)



from typing import TypeVar

from orwynn.base.model import Model
from orwynn.sql.table import Table

TTable = TypeVar("TTable", bound=Table)
ConvertedModel = TypeVar("ConvertedModel", bound=Model)
# Contains many converted models
ListedConvertedModel = TypeVar("ListedConvertedModel", bound=Model)



from collections.abc import Awaitable
from typing import Callable

from orwynn.websocket.websocket import Websocket

# Callable function which returns an awaitable, see:
#   https://stackoverflow.com/a/59177557/14748231
WebsocketNextCall = Callable[[Websocket], Awaitable[None]]



from orwynn.base.module import Module
from orwynn.base.module.errors import CircularDependencyError
from orwynn.utils import validation
from orwynn.utils.fmt import format_chain


class ModuleCollector:
    """Collects all modules starting from root module.

    What is checked:
    - Dependency circular errors for modules
    - Modules self imports
    - Imports of RootModule

    Attributes:
        root_module:
            Root module of the application.
        global_modules (optional):
            Modules to be imported globally for each encountered module.
            Empty by default
    """
    def __init__(
        self,
        root_module: Module,
        *,
        global_modules: list[Module] | None = None
    ) -> None:
        validation.validate(root_module, Module)
        if not global_modules:
            global_modules = []
        validation.validate_each(
            global_modules,
            Module,
            expected_sequence_type=list
        )

        self.__root_module: Module = root_module
        self.__global_modules: list[Module] = global_modules

        self.__collected_modules: list[Module] = []
        self.__chain: list[Module] = []
        # Start traversing from the root module
        self.__traverse(self.__root_module)

    @property
    def collected_modules(self) -> list[Module]:
        return self.__collected_modules.copy()

    def __traverse(
        self,
        init_module: Module
    ) -> None:
        # Just recursively collects all modules, but memorizes every module
        # traversed to find circular dependency errors.

        if init_module in self.__chain:
            raise CircularDependencyError(
                # Failed module is added second time to the chain for
                # error descriptiveness
                f"{init_module} occured twice in dependency chain"
                f" {format_chain(self.__chain + [init_module])}"
            )
        self.__chain.append(init_module)

        if init_module not in self.__collected_modules:
            self.__collected_modules.append(init_module)

            # Add all globally available modules to the current's module
            # imports, and only then continue to inspect the imports.
            #
            # But don't add global modules as imports to another global
            # modules :-)
            if init_module not in self.__global_modules:
                init_module._fw_add_imports(*self.__global_modules)

            if init_module._imports:
                if init_module in init_module._imports:
                    raise CircularDependencyError(
                        f"{init_module} imports self"
                    )
                for m in init_module._imports:
                    # And now start traversing but for an imported module
                    # recursively
                    self.__traverse(m)

        # On blocking case remove recently added module since we don't want
        # this module to appear in other branch, e.g.:
        #   A -> B (blocking: call chain.pop())-> C; A -> B -> D -> C;
        # If C hadn't been removed at the first iteration, we would have got a
        # circular error.
        self.__chain.pop()



from antievil import UnsupportedError

__all__ = [
    "UnsupportedError"
]

class DeprecatedFeatureError(Exception):
    """
    Feature is deprecated.
    """
    def __init__(
        self,
        *,
        deprecated_feature: str,
        use_instead: str
    ):
        message: str = \
            f"feature <{deprecated_feature}> is deprecated," \
            f" use {use_instead} instead"
        super().__init__(message)



from typing import Any

from antievil import TypeConversionError
from pymongo.errors import DuplicateKeyError as PymongoDuplicateKeyError


class UnsetIdMongoError(Exception):
    """
    Id of created Mongo document should be set.
    """
    def __init__(self) -> None:
        super().__init__(
            "id of a created Mongo document should be set",
        )


class MongoTypeConversionError(TypeConversionError):
    """
    Cannot convert an input type to a mongo-compatible type.
    """
    def __init__(
        self,
        *,
        t: type,
    ):
        super().__init__(
            t1=t,
            reason="cannot convert to a mongo-compatible type",
        )


class DuplicateKeyError(Exception):
    def __init__(
        self,
        message: str = "",
        original_error: PymongoDuplicateKeyError | None = None
    ) -> None:
        if not message and original_error:
            message = "; ".join([str(x) for x in original_error.args])
        super().__init__(message)


class DocumentUpdateError(Exception):
    pass


class UnsupportedQueryTypeError(Exception):
    """
    Type of given query is unsupported.
    """
    def __init__(
        self,
        *,
        key: str,
        unsupported_value: Any
    ) -> None:
        message: str = \
            f"query for key <{key}> has an unsupported value" \
            f" <{unsupported_value}>"
        super().__init__(message)



from enum import Enum
from typing import get_args

from orwynn.utils import validation

ErrorCode = int | str | Enum
_ErrorCodeTypes: list[type] = list(get_args(ErrorCode))


def get_error_code(err: Exception) -> ErrorCode:
    """
    Returns an error code if it exists.

    Note that error code should be in class-attribute CODE of the given
    exception and be type of int, str or Enum.

    Args:
        err:
            Exception to get the code from.

    Returns:
        Error code.
    """
    try:
        error_code: ErrorCode = err.CODE  # type: ignore
    except AttributeError as stack_err:
        raise AttributeError(
            f"no specified CODE for error {err}"
        ) from stack_err

    try:
        validation.validate(
            error_code, _ErrorCodeTypes
        )
    except validation.ValidationError as stack_err:
        raise validation.ValidationError(
            f"error code type should be one of {_ErrorCodeTypes}"
        ) from stack_err

    return error_code



from orwynn.base.error.code import ErrorCode, get_error_code

from .errors import MalfunctionError
from .utils import (
    find_detailed_class_for_exception,
)
from .valueschema import ErrorValueSchema
